import contextlib
rWYVM=None
rWYVS=isinstance
rWYVN=list
rWYVH=all
rWYVb=dict
rWYVU=tuple
rWYVw=set
rWYVd=True
rWYVy=type
rWYVK=Exception
rWYVI=str
rWYVs=getattr
rWYVz=len
rWYVk=range
rWYVF=map
rWYVA=bool
rWYVc=False
rWYVR=open
import inspect
import json
import logging
import os
import pickle
import sqlite3
from typing import Any,Dict,Set,Type
from localstack.utils.common import ArbitraryAccessObj
from moto.s3.models import FakeBucket
from moto.sqs.models import Queue
from localstack_ext.bootstrap.state_utils import(check_already_visited,get_object_dict,is_composite_type)
LOG=logging.getLogger(__name__)
DDB_PREDEF_TABLES=("dm","cf","sm","ss","tr","us")
def _merge_helper_three_way(current,injecting,common_ancestor,visited:Set=rWYVM):
 if rWYVS(current,rWYVN)and rWYVS(injecting,rWYVN):
  current.extend(injecting)
  return
 if(not is_composite_type(current)or not is_composite_type(injecting)or not is_composite_type(common_ancestor)):
  return
 cycle,visited=check_already_visited(injecting,visited)
 if cycle:
  return
 cur_dict=get_object_dict(current)
 inj_dict=get_object_dict(injecting)
 ancestor_dict=get_object_dict(common_ancestor)
 for key in cur_dict.keys()&inj_dict.keys()&ancestor_dict.keys():
  if rWYVH(rWYVS(d.get(key),rWYVb)for d in(cur_dict,inj_dict,ancestor_dict)):
   cur_dict[key]=inj_dict[key]=ancestor_dict[key]=_merge_helper_three_way(cur_dict[key],inj_dict[key],ancestor_dict[key])
 for d in cur_dict,inj_dict,ancestor_dict:
  for k,v in d.items():
   if rWYVS(v,rWYVb):
    d[k]=rWYVU(v.items())
 set_cur,set_inj,set_ancestor=(rWYVw((k,pickle.dumps(v))for k,v in d.items())for d in(cur_dict,inj_dict,ancestor_dict))
 for k in rWYVw(k for k,_ in set_cur^set_ancestor)&rWYVw(k for k,_ in set_inj^set_ancestor):
  if cur_dict.get(k)!=inj_dict.get(k)or(k in cur_dict)^(k in inj_dict):
   LOG.debug("Found update-update conflict")
   cur_dict[k]=inj_dict[k]
 d=rWYVb(set_cur&set_inj&set_ancestor|set_cur-set_ancestor|set_inj-set_ancestor)
 for k,v in d.items():
  v=pickle.loads(v)
  if rWYVS(v,rWYVU):
   d[k]=rWYVb(v)
  else:
   d[k]=v
 return d
def _merge_helper(current,injecting,merge_strategy=rWYVM,visited:Set=rWYVM):
 if rWYVS(current,rWYVN)and rWYVS(injecting,rWYVN):
  current.extend(injecting)
  return
 if not is_composite_type(current)or not is_composite_type(injecting):
  return
 cycle,visited=check_already_visited(injecting,visited)
 if cycle:
  return
 cur_dict=get_object_dict(current)
 inj_dict=get_object_dict(injecting)
 for field_name,inj_field_value in inj_dict.items():
  cur_field_value=cur_dict.get(field_name)
  if cur_field_value is not rWYVM:
   if is_composite_type(cur_field_value):
    _merge_helper(cur_field_value,inj_field_value,merge_strategy=merge_strategy,visited=visited)
   elif cur_field_value!=inj_field_value:
    LOG.debug("Overwriting existing value with new state: '%s' <> '%s'"%(cur_field_value,inj_field_value))
    cur_dict[field_name]=inj_field_value
  else:
   cur_dict[field_name]=inj_field_value
 return cur_dict
def merge_object_state(current,injecting,common_ancestor=rWYVM,merge_strategy=rWYVM):
 if not current or not injecting:
  return current
 is_special_case=handle_special_case(current,injecting)
 if is_special_case:
  return current
 if common_ancestor:
  _merge_helper_three_way(current,injecting,common_ancestor)
 else:
  _merge_helper(current,injecting)
 add_missing_attributes(current)
 return current
def handle_special_case(current,injecting):
 if rWYVS(injecting,Queue):
  current.queues[injecting.name]=injecting
  return rWYVd
 elif rWYVS(injecting,FakeBucket):
  current["global"].buckets[injecting.name]=injecting
  return rWYVd
def add_missing_attributes(obj:Any,safe:rWYVA=rWYVd,visited:Set=rWYVM):
 try:
  obj_dict=get_object_dict(obj)
  if obj_dict is rWYVM:
   return
  cycle,visited=check_already_visited(obj,visited)
  if cycle:
   return
  for attr_value in obj_dict.values():
   add_missing_attributes(attr_value,safe=safe,visited=visited)
  class_inst_attrs=infer_class_attributes(rWYVy(obj))
  for key,value in class_inst_attrs.items():
   if key not in obj_dict:
    LOG.debug("Add missing attribute '%s' to state object of type %s"%(key,rWYVy(obj)))
    obj_dict[key]=value
 except rWYVK as e:
  if not safe:
   raise
  LOG.warning("Unable to add missing attributes to persistence state object %s: %s",(obj,e))
def infer_class_attributes(clazz:Type)->Dict[rWYVI,Any]:
 if clazz in[rWYVN,rWYVb]or not inspect.isclass(clazz)or clazz.__name__=="function":
  return{}
 constructor=rWYVs(clazz,"__init__",rWYVM)
 if not constructor:
  return{}
 try:
  sig_args=inspect.getfullargspec(constructor)
  def get_default_arg_value(arg_name,arg_index=-1):
   arg_defaults=sig_args.defaults or[]
   num_non_default_args=rWYVz(sig_args.args or[])-rWYVz(arg_defaults)
   offset=arg_index-num_non_default_args
   if offset in rWYVk(rWYVz(arg_defaults)):
    return arg_defaults[offset]
   kwargs_defaults=sig_args.kwonlydefaults or{}
   if arg_name in kwargs_defaults:
    return kwargs_defaults[arg_name]
   return ArbitraryAccessObj()
  args=[]
  kwargs={}
  for arg_idx in rWYVk(1,rWYVz(sig_args.args)):
   args.append(get_default_arg_value(sig_args.args[arg_idx],arg_index=arg_idx))
  for arg in sig_args.kwonlyargs:
   kwargs[arg]=get_default_arg_value(arg)
  instance=clazz(*args,**kwargs)
  result=rWYVb(instance.__dict__)
  return result
 except rWYVK:
  return{}
def merge_sqllite_dbs(file_dest:rWYVI,file_src:rWYVI)->rWYVM:
 def _merge_table(table_name:rWYVI,cursor_a,cursor_b)->rWYVM:
  tmp_table_name=f"'{table_name}_new'"
  table_name=f"'{table_name}'"
  select_query=f"SELECT * FROM {table_name}"
  if table_name=="'cf'":
   return
  schema=rWYVU(rWYVF(lambda x:x[1],cursor_b.execute(f"PRAGMA table_info({table_name})")))
  params=f"({('?,' * len(schema))[:-1]})"
  insert_str=f"INSERT INTO {tmp_table_name} {str(schema)} values {params}"
  if table_name=="'dm'":
   excl_tables=(rWYVI(rWYVN(rWYVF(lambda x:x[0],cursor_a.execute(f"SELECT TableName FROM {table_name}")))).replace("[","(",1).replace("]",")",1))
   select_query+=f"AS O_T WHERE O_T.TableName NOT IN {excl_tables}"
  cursor_a.execute(f"CREATE TABLE IF NOT EXISTS {tmp_table_name} AS SELECT * FROM {table_name}")
  for row in cursor_b.execute(select_query):
   cursor_a.execute(insert_str,row)
  cursor_a.execute(f"DROP TABLE IF EXISTS {table_name}")
  cursor_a.execute(f"ALTER TABLE {tmp_table_name} RENAME TO {table_name}")
 with contextlib.closing(sqlite3.connect(file_dest))as db_src,contextlib.closing(sqlite3.connect(file_src))as db_target:
  cursor_dest=db_src.cursor()
  cursor_src=db_target.cursor()
  table_names=rWYVN(rWYVF(lambda x:x[0],cursor_dest.execute("SELECT name FROM sqlite_master WHERE type='table'")))
  for current_table in table_names:
   try:
    _merge_table(current_table,cursor_dest,cursor_src)
   except sqlite3.OperationalError as e:
    LOG.warning(f"Failed to merge table {current_table}: {e}")
    cursor_dest.execute(f"DROP TABLE IF EXISTS '{current_table}'")
    db_src.rollback()
    return
  db_src.commit()
  LOG.debug(f"Successfully merged db at {file_src} into {file_dest}")
def merge_kinesis_state(path_dest:rWYVI,path_src:rWYVI)->rWYVA:
 state_file="kinesis-data.json"
 datadir_statefile=os.path.join(path_dest,state_file)
 tmp_dir_statefile=os.path.join(path_src,state_file)
 if not os.path.isfile(datadir_statefile):
  LOG.info(f"Could not find statefile in path destination {path_dest}")
  return rWYVc
 if not os.path.isfile(tmp_dir_statefile):
  LOG.info(f"Could not find statefile in path source {path_src}")
  return rWYVc
 with rWYVR(datadir_statefile)as datadir_kinesis_file,rWYVR(tmp_dir_statefile)as tmp_dir_kinesis_file:
  datadir_kinesis_json=json.load(datadir_kinesis_file)
  tmp_dir_kinesis_json=json.load(tmp_dir_kinesis_file)
  datadir_streams=datadir_kinesis_json.get("streams",[])
  tmp_dir_streams=tmp_dir_kinesis_json.get("streams",[])
  if rWYVz(tmp_dir_streams)>0:
   datadir_stream_names=datadir_streams.keys()
   for stream in tmp_dir_streams:
    if stream not in datadir_stream_names:
     datadir_streams[stream]=tmp_dir_streams.get(stream)
     LOG.debug(f"Copied state from stream {stream}")
   with rWYVR(datadir_statefile,"w")as mutated_datadir_kinesis_file:
    mutated_datadir_kinesis_file.write(json.dumps(datadir_kinesis_json))
   return rWYVd
 return rWYVc
# Created by pyminifier (https://github.com/liftoff/pyminifier)



import json
oqCeX=staticmethod
oqCex=classmethod
oqCew=None
oqCeN=True
oqCer=hasattr
oqCeA=Exception
oqCeF=super
from localstack.services.cloudformation.service_models import(REF_ARN_ATTRS,REF_ID_ATTRS,GenericBaseModel)
from localstack.utils.aws import aws_stack
from localstack.utils.common import clone,keys_to_lower,select_attributes,to_str
from localstack_ext.services.appsync import appsync_api
from localstack_ext.services.cloudformation.service_models import LOG
from localstack_ext.utils.aws import aws_utils
class AppSyncResolver(GenericBaseModel):
 @oqCeX
 def cloudformation_type():
  return "AWS::AppSync::Resolver"
 def get_physical_resource_id(self,attribute,**kwargs):
  if attribute in REF_ID_ATTRS:
   return self.props.get("resolverArn")
 def fetch_state(self,stack_name,resources):
  props=self.props
  api_id=props.get("ApiId")
  type_name=props.get("TypeName")
  field_name=props.get("FieldName")
  data_source_name=props.get("DataSourceName")
  return self.fetch_details(api_id,type_name,field_name,data_source_name)
 @oqCex
 def fetch_details(cls,api_id,type_name,field_name,data_source_name):
  client=aws_stack.connect_to_service("appsync")
  resolvers=client.list_resolvers(apiId=api_id,typeName=type_name)["resolvers"]
  result=[r for r in resolvers if r.get("fieldName")==field_name and r.get("dataSourceName")==data_source_name]
  return(result or[oqCew])[0]
 @oqCeX
 def get_deploy_templates():
  def get_resolver_params(params,**kwargs):
   params=keys_to_lower(clone(params))
   req_template=params.pop("requestMappingTemplateS3Location",oqCew)
   res_template=params.pop("responseMappingTemplateS3Location",oqCew)
   if req_template:
    tmpl=aws_utils.download_s3_object(req_template,as_str=oqCeN)
    if tmpl:
     params["requestMappingTemplate"]=tmpl
   if res_template:
    tmpl=aws_utils.download_s3_object(res_template,as_str=oqCeN)
    if tmpl:
     params["responseMappingTemplate"]=tmpl
   return params
  return{"create":{"function":"create_resolver","parameters":get_resolver_params}}
class GraphQLSchema(GenericBaseModel):
 @oqCeX
 def cloudformation_type():
  return "AWS::AppSync::GraphQLSchema"
 def get_physical_resource_id(self,attribute,**kwargs):
  return "%s-GraphQLSchema"%self.props.get("ApiId")
 @oqCex
 def fetch_details(cls,api_id,definition=oqCew,s3_location=oqCew):
  client=aws_stack.connect_to_service("appsync")
  try:
   schema=client.get_introspection_schema(apiId=api_id,format="SDL")["schema"]
   schema=schema.read()if oqCer(schema,"read")else schema
   schema=to_str(schema)
   if schema.startswith("{"):
    schema=json.loads(schema)["schema"]
   if definition:
    if to_str(definition).strip()==to_str(schema).strip():
     return{"schema":definition}
   elif s3_location:
    LOG.debug("Getting GraphQL schema definition from S3 not yet implemented")
    return{"schema":definition}
  except oqCeA:
   pass
class AppSyncApiKey(GenericBaseModel):
 @oqCeX
 def cloudformation_type():
  return "AWS::AppSync::ApiKey"
 def get_cfn_attribute(self,attribute_name):
  props=self.props
  if attribute_name=="ApiKey":
   return "key-%s"%props["ApiId"]
  if attribute_name in REF_ID_ATTRS+REF_ARN_ATTRS:
   return self.get_physical_resource_id()
  return oqCeF(AppSyncApiKey,self).get_cfn_attribute(attribute_name)
 def get_physical_resource_id(self,attribute=oqCew,**kwargs):
  props=self.props
  key_id=props.get("ApiKeyId")or props.get("id")
  if not key_id or not props.get("ApiId"):
   return
  return aws_utils.appsync_api_key_arn(props.get("ApiId"),key_id)
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("appsync")
  props=self.props
  api_id=props.get("ApiId")
  api_id=self.resolve_refs_recursively(stack_name,api_id,resources)
  keys=client.list_api_keys(apiId=api_id).get("apiKeys",[])
  result=[k for k in keys if k.get("description")==props.get("Description")and props.get("Expires")in[oqCew,k.get("expires")]]
  return(result or[oqCew])[0]
 @oqCex
 def get_deploy_templates(cls):
  def _create_params(_,resources,resource_id,**kwargs):
   resource=cls(resources[resource_id])
   res_props=resource.properties
   if not res_props.get("Description"):
    api_id=resource.resolve_refs_recursively(kwargs.get("stack_name"),res_props.get("ApiId"),resources)
    res_props["Description"]=f"Generated API key for AppSync API ID {api_id}"
   result=keys_to_lower(select_attributes(res_props,["ApiId","Description","Expires"]))
   return result
  return{"create":{"function":"create_api_key","parameters":_create_params}}
class AppSyncDataSource(GenericBaseModel):
 @oqCeX
 def cloudformation_type():
  return "AWS::AppSync::DataSource"
 @oqCeX
 def get_deploy_templates():
  def lambda_get_datasource_config(params,**kwargs):
   params=keys_to_lower(clone(params))
   if params.get("dynamoDBConfig"):
    params["dynamodbConfig"]=params.pop("dynamoDBConfig")
   return params
  return{"create":{"function":"create_data_source","parameters":lambda_get_datasource_config}}
 def get_cfn_attribute(self,attribute_name):
  if attribute_name in REF_ARN_ATTRS+["DataSourceArn"]:
   return self.props.get("dataSourceArn")
  if attribute_name=="Name":
   return self.props.get("name")
  return oqCeF(AppSyncDataSource,self).get_cfn_attribute(attribute_name)
 def get_physical_resource_id(self,attribute=oqCew,**kwargs):
  return self.props.get("dataSourceArn")
 def fetch_state(self,stack_name,resources):
  props=self.props
  client=aws_stack.connect_to_service("appsync")
  ds_name=props.get("Name")or self.resource_id
  ds_name=self.resolve_refs_recursively(stack_name,ds_name,resources)
  api_id=self.resolve_refs_recursively(stack_name,props.get("ApiId"),resources)
  sources=client.list_data_sources(apiId=api_id)["dataSources"]
  return([s for s in sources if s["name"]==ds_name]or[oqCew])[0]
class AppSyncFunctionConfig(GenericBaseModel):
 @oqCeX
 def cloudformation_type():
  return "AWS::AppSync::FunctionConfiguration"
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("appsync")
  props=self.props
  api_id=self.resolve_refs_recursively(stack_name,props["ApiId"],resources)
  name=self.resolve_refs_recursively(stack_name,props["Name"],resources)
  result=client.list_functions(apiId=api_id).get("functions",[])
  result=[f for f in result if f.get("name")==name]
  return(result or[oqCew])[0]
 def get_cfn_attribute(self,attribute_name):
  props=self.props
  if attribute_name=="FunctionId":
   return props.get("functionArn","").split("functions/")[-1]
  if attribute_name=="FunctionArn":
   return props.get("functionArn")
  if attribute_name=="Name":
   return props.get("name")
  if attribute_name=="DataSourceName":
   return props.get("dataSourceName")
  return oqCeF(AppSyncFunctionConfig,self).get_cfn_attribute(attribute_name)
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("functionArn")
 @oqCeX
 def get_deploy_templates():
  def func_params(params,**kwargs):
   sys_cfg=params.get("SyncConfig")
   if sys_cfg:
    sys_cfg=keys_to_lower(clone(sys_cfg))
   result={"apiId":"ApiId","name":"Name","description":"Description","dataSourceName":"DataSourceName","requestMappingTemplate":"RequestMappingTemplate","responseMappingTemplate":"ResponseMappingTemplate","functionVersion":"FunctionVersion","syncConfig":sys_cfg}
   result={k:params.get(v)for k,v in result.items()if params.get(v)}
   return result
  return{"create":{"function":"create_function","parameters":func_params}}
class GraphQLAPI(GenericBaseModel):
 @oqCeX
 def cloudformation_type():
  return "AWS::AppSync::GraphQLApi"
 def get_cfn_attribute(self,attribute_name):
  props=self.props
  if not props.get("apiId"):
   details=self.fetch_details(props.get("Name"))
   if not details:
    raise oqCeA('Unable to find GraphQL API named "%s"'%props.get("Name"))
   props["apiId"]=details["apiId"]
  if attribute_name=="ApiId":
   return props.get("apiId")
  if attribute_name in REF_ARN_ATTRS:
   return self.get_physical_resource_id(attribute_name)
  if attribute_name=="GraphQLUrl":
   return appsync_api.get_graphql_url(props["apiId"],use_domain_name=oqCeN)
  return oqCeF(GraphQLAPI,self).get_cfn_attribute(attribute_name)
 def get_physical_resource_id(self,attribute,**kwargs):
  props=self.props
  return props.get("apiId")and aws_utils.appsync_api_arn(props["apiId"])
 @oqCex
 def fetch_details(cls,api_name):
  client=aws_stack.connect_to_service("appsync")
  apis=client.list_graphql_apis()["graphqlApis"]
  return([a for a in apis if a["name"]==api_name]or[oqCew])[0]
# Created by pyminifier (https://github.com/liftoff/pyminifier)



from localstack.services.cloudformation.service_models import REF_ID_ATTRS,GenericBaseModel
tqBNF=staticmethod
tqBNn=super
class KinesisAnalyticsApplicationOutput(GenericBaseModel):
 @tqBNF
 def cloudformation_type():
  return "AWS::KinesisAnalytics::ApplicationOutput"
 def get_cfn_attribute(self,attribute_name):
  if attribute_name in REF_ID_ATTRS:
   app_name=self.props.get("ApplicationName")
   output_name=self.props.get("Output",{}).get("Name")
   if app_name and output_name:
    return "%s!%s"%(app_name,output_name)
  return tqBNn(KinesisAnalyticsApplicationOutput,self).get_cfn_attribute(attribute_name)
class KinesisAnalyticsApplication(GenericBaseModel):
 @tqBNF
 def cloudformation_type():
  return "AWS::KinesisAnalytics::Application"
 def get_cfn_attribute(self,attribute_name):
  if attribute_name in REF_ID_ATTRS:
   app_name=self.props.get("ApplicationName")
   if app_name:
    return app_name
  return tqBNn(KinesisAnalyticsApplication,self).get_cfn_attribute(attribute_name)
# Created by pyminifier (https://github.com/liftoff/pyminifier)



from localstack.services.cloudformation.deployment_utils import param_json_to_str
KGMwW=staticmethod
KGMwx=None
KGMwE=classmethod
KGMwg=str
from localstack.services.cloudformation.service_models import REF_ID_ATTRS,GenericBaseModel
from localstack.utils.aws import aws_stack
from localstack.utils.common import clone,select_attributes
from localstack_ext.services.cloudformation.service_models import LOG
class EC2VPCEndpoint(GenericBaseModel):
 @KGMwW
 def cloudformation_type():
  return "AWS::EC2::VPCEndpoint"
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("ec2")
  result=client.describe_vpc_endpoints(Filters=[{"Name":"service-name","Values":[self.props["ServiceName"]]},{"Name":"vpc-id","Values":[self.props["VpcId"]]}])
  result=result["VpcEndpoints"]
  return(result or[KGMwx])[0]
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("VpcEndpointId")
 @KGMwW
 def get_deploy_templates():
  return{"create":{"function":"create_vpc_endpoint","parameters":{"ServiceName":"ServiceName","PolicyDocument":param_json_to_str("PolicyDocument"),"VpcId":"VpcId","SecurityGroupIds":"SecurityGroupIds","SubnetIds":"SubnetIds","RouteTableIds":"RouteTableIds","VpcEndpointType":"VpcEndpointType","PrivateDnsEnabled":"PrivateDnsEnabled"}}}
class EC2ElasticIP(GenericBaseModel):
 @KGMwW
 def cloudformation_type():
  return "AWS::EC2::EIP"
 def get_physical_resource_id(self,attribute=KGMwx,**kwargs):
  return self.props.get("PublicIp")
 def fetch_state(self,stack_name,resources):
  if self.get_physical_resource_id():
   return self.state
 @KGMwE
 def get_deploy_templates(cls):
  def allocate_address(resource_id,resources,resource_type,func,stack_name,*args,**kwargs):
   resource=cls(resources[resource_id])
   resource.fetch_and_update_state(stack_name,resources)
   client=aws_stack.connect_to_service("ec2")
   kwargs=select_attributes(resource.props,["Domain","PublicIpv4Pool"])
   result=client.allocate_address(**kwargs)
   resource.state.update(result)
   return result
  return{"create":{"function":allocate_address},"delete":{"function":"release_address","parameters":["PublicIp","AllocationId"]}}
class SubnetRouteTableAssociation(GenericBaseModel):
 @KGMwW
 def cloudformation_type():
  return "AWS::EC2::SubnetRouteTableAssociation"
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("RouteTableAssociationId")
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("ec2")
  props=self.props
  table_id=self.resolve_refs_recursively(stack_name,props["RouteTableId"],resources)
  subnet_id=self.resolve_refs_recursively(stack_name,props["SubnetId"],resources)
  tables=client.describe_route_tables(RouteTableIds=[table_id])["RouteTables"]
  tables=[t for t in tables if t["RouteTableId"]==table_id]
  if tables:
   assocs=[a for a in tables[0].get("Associations",[])if a["SubnetId"]==subnet_id]
   return(assocs or[KGMwx])[0]
 def get_deploy_templates():
  return{"create":{"function":"associate_route_table","parameters":{"RouteTableId":"RouteTableId","SubnetId":"SubnetId"}},"delete":{"function":"disassociate_route_table","parameters":{"AssociationId":"PhysicalResourceId"}}}
class SecurityGroupInOrEgress(GenericBaseModel):
 def get_physical_resource_id(self,attribute,**kwargs):
  if attribute in REF_ID_ATTRS:
   props=self.props
   res_id="%s_%s_%s"%(props.get("IpProtocol"),props.get("FromPort"),props.get("ToPort"))
   return res_id
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("ec2")
  rp=self.props
  kwargs=({"GroupNames":[rp.get("GroupName")]}if rp.get("GroupName")else{"GroupIds":[rp["GroupId"]]})
  result=client.describe_security_groups(**kwargs)["SecurityGroups"]
  src_grp_name=rp.get("SourceSecurityGroupName")
  src_grp_id=rp.get("SourceSecurityGroupId")
  dst_grp_id=rp.get("DestinationSecurityGroupId")
  if result:
   perms=result[0].get("IpPermissions" if self.is_ingress()else "IpPermissionsEgress")
   result=[]
   for perm in perms:
    if KGMwg(perm["IpProtocol"])!=KGMwg(rp["IpProtocol"]):
     continue
    if perm.get("FromPort")!=rp.get("FromPort")or perm.get("ToPort")!=rp.get("ToPort"):
     continue
    if not self.is_ingress():
     return perm
    groups=perm.get("UserIdGroupPairs",[])
    groups=[g for g in groups if g.get("GroupId")in[src_grp_id,dst_grp_id]or g.get("GroupName")==src_grp_name]
    if groups:
     return perm
 @KGMwE
 def is_ingress(cls):
  return "Ingress" in cls.cloudformation_type()
 @KGMwE
 def get_deploy_templates(cls):
  def create_params(params,**kwargs):
   result=clone(params)
   source_group_name=result.get("SourceSecurityGroupName")
   source_group_id=result.pop("SourceSecurityGroupId",KGMwx)
   vpc_id=KGMwx
   if cls.is_ingress()and source_group_id and not source_group_name:
    client=aws_stack.connect_to_service("ec2")
    groups=client.describe_security_groups(GroupIds=[source_group_id])["SecurityGroups"]
    result["SourceSecurityGroupName"]=groups and groups[0]["GroupName"]or KGMwx
    vpc_id=groups and groups[0].get("VpcId")or KGMwx
   dst_group_id=result.pop("DestinationSecurityGroupId",KGMwx)
   if not cls.is_ingress()and not dst_group_id:
    LOG.info("TODO: Add support for DestinationPrefixListId for %s"%cls.cloudformation_type())
   if result.get("IpProtocol"):
    result["IpProtocol"]=KGMwg(result.get("IpProtocol"))
   description=result.pop("Description",KGMwx)
   cidr_ipv6=result.pop("CidrIpv6",KGMwx)
   if cidr_ipv6 or vpc_id:
    cidr_ip=result.get("CidrIp")
    cidr_ip=cidr_ip or("127.0.0.1/32" if description else cidr_ip)
    ipv6_range={"CidrIpv6":cidr_ipv6,"Description":description}
    ip_range={"CidrIp":cidr_ip,"Description":description}
    groups=[]
    if source_group_id:
     groups.append({"GroupId":source_group_id,"GroupName":source_group_name,"Description":description,"VpcId":vpc_id})
    if dst_group_id:
     groups.append({"GroupId":dst_group_id,"Description":description,"VpcId":vpc_id})
    ip_perm={"IpProtocol":result.get("IpProtocol"),"UserIdGroupPairs":groups,"FromPort":result.get("FromPort"),"ToPort":result.get("ToPort")}
    ip_perm["IpRanges"]=cidr_ip and[ip_range]
    ip_perm["Ipv6Ranges"]=cidr_ipv6 and[ipv6_range]
    result["IpPermissions"]=[ip_perm]
   else:
    LOG.debug('Neither "VpcId" nor "CidrIpv6" found in CF params: %s'%params)
   return result
  func_name=("authorize_security_group_ingress" if cls.is_ingress()else "authorize_security_group_egress")
  return{"create":{"function":func_name,"parameters":create_params}}
class SecurityGroupEgress(SecurityGroupInOrEgress):
 @KGMwW
 def cloudformation_type():
  return "AWS::EC2::SecurityGroupEgress"
class SecurityGroupIngress(SecurityGroupInOrEgress):
 @KGMwW
 def cloudformation_type():
  return "AWS::EC2::SecurityGroupIngress"
# Created by pyminifier (https://github.com/liftoff/pyminifier)



from localstack.services.cloudformation.models import awslambda
czWLh=staticmethod
czWLI=super
czWLG=None
czWLP=int
czWLN=len
czWLR=classmethod
from localstack.services.cloudformation.service_models import REF_ARN_ATTRS,GenericBaseModel
from localstack.utils.aws import aws_stack
from localstack.utils.common import select_attributes,short_uid
class LambdaLayerVersion(GenericBaseModel):
 @czWLh
 def cloudformation_type():
  return "AWS::Lambda::LayerVersion"
 def get_cfn_attribute(self,attribute_name):
  if attribute_name in REF_ARN_ATTRS:
   version="$LATEST"
   return aws_stack.lambda_layer_arn(self.props["LayerName"],version=version)
  return czWLI(LambdaLayerVersion,self).get_cfn_attribute(attribute_name)
 def fetch_state(self,stack_name,resources):
  layer_name=self.resolve_refs_recursively(stack_name,self.props.get("LayerName"),resources)
  client=aws_stack.connect_to_service("lambda")
  layers=client.list_layer_versions(LayerName=layer_name).get("LayerVersions",[])
  return layers[-1]if layers else czWLG
 @czWLh
 def get_deploy_templates():
  return{"create":{"function":"publish_layer_version"}}
class LambdaLayerVersionPermission(awslambda.LambdaPermission):
 @czWLh
 def cloudformation_type():
  return "AWS::Lambda::LayerVersionPermission"
 def fetch_state(self,stack_name,resources):
  props=self.props
  props["LayerVersionArn"]=self.resolve_refs_recursively(stack_name,props["LayerVersionArn"],resources)
  layer_name,version_number=self.layer_name_and_version(props)
  layer_arn=aws_stack.lambda_layer_arn(layer_name)
  layer_arn_qualified="%s:%s"%(layer_arn,version_number)
  result=self.do_fetch_state(layer_name,layer_arn_qualified)
  return result
 @czWLh
 def layer_name_and_version(params):
  layer_arn=params.get("LayerVersionArn","")
  parts=layer_arn.split(":")
  layer_name=parts[6]if ":" in layer_arn else layer_arn
  version_number=czWLP(parts[7]if czWLN(parts)>7 else 1)
  return layer_name,version_number
 @czWLR
 def get_deploy_templates(cls):
  def layer_permission_params(params,**kwargs):
   layer_name,version_number=cls.layer_name_and_version(params)
   result=select_attributes(params,["Action","Principal"])
   result["StatementId"]=short_uid()
   result["LayerName"]=layer_name
   result["VersionNumber"]=version_number
   return result
  return{"create":{"function":"add_layer_version_permission","parameters":layer_permission_params}}
class LambdaAlias(awslambda.LambdaPermission):
 @czWLh
 def cloudformation_type():
  return "AWS::Lambda::Alias"
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("lambda")
  props=self.props
  func_name=self.resolve_refs_recursively(stack_name,props.get("FunctionName"),resources)
  func_version=self.resolve_refs_recursively(stack_name,props.get("FunctionVersion"),resources)
  layers=client.list_aliases(FunctionName=func_name)["Aliases"]
  result=[lr for lr in layers if func_version in(czWLG,lr.get("FunctionVersion"))]
  return(result or[czWLG])[0]
 def get_physical_resource_id(self,attribute,**kwargs):
  repo_name=self.props.get("AliasArn")
  return repo_name
 @czWLR
 def get_deploy_templates(cls):
  return{"create":{"function":"create_alias","parameters":["Description","FunctionName","FunctionVersion","Name"]}}
# Created by pyminifier (https://github.com/liftoff/pyminifier)



from localstack.services.cloudformation.service_models import GenericBaseModel
OHuLs=staticmethod
OHuLx=None
OHuLW=False
from localstack.utils.aws import aws_stack
from localstack_ext.utils.aws import aws_utils
class ECRRepository(GenericBaseModel):
 @OHuLs
 def cloudformation_type():
  return "AWS::ECR::Repository"
 def get_physical_resource_id(self,attribute,**kwargs):
  repo_name=self.props.get("RepositoryName")
  if attribute=="Arn":
   return aws_utils.get_ecr_repository_arn(repo_name)
  return repo_name
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("ecr")
  props=self.props
  repo_name=props.get("RepositoryName")
  registry_id=props.get("LifecyclePolicy",{}).get("RegistryId")
  kwargs={"registryId":registry_id}if registry_id else{}
  result=client.describe_repositories(repositoryNames=[repo_name],**kwargs).get("repositories",[])
  return(result or[OHuLx])[0]
 @OHuLs
 def get_deploy_templates():
  def _create_params(params,**kwargs):
   encryption_type=params.get("EncryptionConfiguration",{}).get("EncryptionType","AES-256")
   response={"repositoryName":params["RepositoryName"],"tags":params.get("Tags",[]),"imageTagMutability":params.get("ImageTagMutability","MUTABLE"),"imageScanningConfiguration":{"scanOnPush":params.get("ImageScanningConfiguration",{}).get("ScanOnPush",OHuLW)},"encryptionConfiguration":{"encryptionType":encryption_type}}
   if(encryption_type=="KMS" and params.get("EncryptionConfiguration",{}).get("KmsKey")is not OHuLx):
    response["encryptionConfiguration"]["kmsKey"]=params["EncryptionConfiguration"]["KmsKey"]
   return response
  result={"create":{"function":"create_repository","parameters":_create_params},"delete":{"function":"delete_repository","parameters":{"repositoryName":"RepositoryName"}}}
  return result
# Created by pyminifier (https://github.com/liftoff/pyminifier)



import logging
pYiMN=str
pYiMt=False
pYiMP=open
pYiMV=Exception
pYiMa=None
pYiMd=isinstance
pYiMw=dict
pYiMG=list
pYiMf=len
import os
import shutil
import zipfile
from typing import Dict
from localstack.utils.common import merge_recursive,new_tmp_dir,rm_rf
from localstack_ext.bootstrap.cpvcs.constants import COMPRESSION_FORMAT
from localstack_ext.bootstrap.cpvcs.obj_storage import default_storage as object_storage
from localstack_ext.bootstrap.cpvcs.utils.common import CPVCSConfigContext,config_context
from localstack_ext.bootstrap.state_utils import(check_already_visited,get_object_dict,load_persisted_object)
LOG=logging.getLogger(__name__)
def extract_meta_and_state_archives(meta_archives:Dict[pYiMN,pYiMN],state_archives:Dict[pYiMN,pYiMN]):
 for product_space_archive in[meta_archives,state_archives]:
  for version_no,archive in product_space_archive.items():
   with zipfile.ZipFile(archive)as meta_zip:
    if product_space_archive==meta_archives:
     archive_dest=config_context.get_version_meta_archive_path(version_no=version_no,with_format=pYiMt)
    else:
     archive_dest=config_context.get_version_state_archive_path(version_no=version_no,with_format=pYiMt)
    meta_zip.extractall(archive_dest)
    shutil.make_archive(base_name=archive_dest,format=COMPRESSION_FORMAT,root_dir=archive_dest)
    rm_rf(archive_dest)
    rm_rf(archive)
    LOG.debug(f"Successfully extracted archive {product_space_archive} for version {version_no}")
def register_remote(remote_info:Dict[pYiMN,pYiMN]):
 if config_context.is_remotly_managed():
  LOG.warning("Pod is already remotely managed")
  return
 with pYiMP(config_context.get_remote_info_path(),"w")as fp:
  storage_uuid=remote_info.get("storage_uuid")
  qualifying_name=remote_info.get("qualifying_name")
  fp.write(f"storage_uuid={storage_uuid}\n")
  fp.write(f"qualifying_name={qualifying_name}\n")
def merge_version_space(version_space_archive):
 remote_version_space_dir=new_tmp_dir()
 remote_config_context=CPVCSConfigContext(pod_root_dir=remote_version_space_dir)
 with zipfile.ZipFile(version_space_archive)as version_space_zip:
  version_space_zip.extractall(remote_config_context.get_pod_root_dir())
 shutil.copy(remote_config_context.get_known_ver_path(),config_context.get_known_ver_path())
 shutil.copy(remote_config_context.get_max_ver_path(),config_context.get_max_ver_path())
 remote_rev_obj_store_path=remote_config_context.get_rev_obj_store_path()
 local_rev_obj_store_path=config_context.get_rev_obj_store_path()
 for revision_file in os.listdir(remote_rev_obj_store_path):
  remote_revision_file_path=os.path.join(remote_rev_obj_store_path,revision_file)
  local_revision_file_path=os.path.join(local_rev_obj_store_path,revision_file)
  shutil.copy(remote_revision_file_path,local_revision_file_path)
 for version_ref_file in os.listdir(remote_config_context.get_ver_refs_path()):
  remote_version_ref_file_path=remote_config_context.get_version_ref_file_path(version_ref_file)
  local_version_ref_file_path=config_context.get_version_ref_file_path(version_ref_file)
  with pYiMP(remote_version_ref_file_path,"r")as fp:
   key=fp.readline().strip()
  if os.path.isfile(local_version_ref_file_path):
   object_storage.merge_remote_into_local_version(remote_config_context.get_ver_obj_store_path(),key)
  else:
   remote_version_file_path=os.path.join(remote_config_context.get_ver_obj_store_path(),key)
   local_version_file_path=os.path.join(config_context.get_ver_obj_store_path(),key)
   shutil.copy(remote_version_ref_file_path,local_version_ref_file_path)
   shutil.copy(remote_version_file_path,local_version_file_path)
 rm_rf(remote_version_space_dir)
 rm_rf(version_space_archive)
def _metadata_create_func(**kwargs):
 try:
  dir_name=kwargs.get("dir_name")
  file_name=kwargs.get("fname")
  region=kwargs.get("region")
  service_name=kwargs.get("service_name")
  mutables=kwargs.get("mutables")
  metamodels=mutables[0]
  file_path=os.path.join(dir_name,file_name)
  backend_state=load_persisted_object(file_path)
  service_metamodel=_create_metamodel_helper(backend_state)or{}
  region_metamodels=metamodels[region]=metamodels.get(region)or{}
  region_service_models=region_metamodels[service_name]=(region_metamodels.get(service_name)or{})
  service_info=mutables[1]
  merge_recursive(service_metamodel,region_service_models)
  service_region_info=service_info.setdefault(region,{})
  service_info=service_region_info.setdefault(service_name,{})
  service_info["size"]=service_info.get("size",0)+os.path.getsize(file_path)
 except pYiMV as e:
  LOG.exception(f"Unable to create metamodel for state object {kwargs} : {e}")
def _create_metamodel_helper(obj,width=25,visited:set=pYiMa):
 if obj is pYiMa:
  return obj
 cycle,visited=check_already_visited(obj,visited)
 if cycle:
  return obj
 result=obj=get_object_dict(obj)or obj
 if pYiMd(obj,pYiMw):
  result=pYiMw(result)
  for field_name,field_value in result.items():
   result[field_name]=_create_metamodel_helper(field_value,width=width,visited=visited)
 elif pYiMd(obj,pYiMG):
  result=[_create_metamodel_helper(o,width=width,visited=visited)for o in obj]
  if pYiMf(result)>width:
   result={"size":pYiMf(result),"items":result[:width]}
 return result
# Created by pyminifier (https://github.com/liftoff/pyminifier)



import inspect
FsTdk=None
FsTdi=set
FsTdW=open
FsTdS=str
FsTdO=bool
FsTdA=False
FsTdq=True
FsTdn=filter
FsTdx=int
FsTdE=Exception
FsTdB=next
FsTdj=sorted
FsTdy=isinstance
FsTdC=dict
FsTdo=getattr
FsTdL=type
import json
import logging
import os
import shutil
import zipfile
from typing import Dict,List,Optional,Set,Tuple
from deepdiff import DeepDiff
from localstack.utils.common import json_safe,mkdir,new_tmp_dir,rm_rf,short_uid
from localstack.utils.testutil import create_zip_file_python
from moto.s3.models import FakeBucket
from moto.sqs.models import Queue
from localstack_ext.bootstrap.cpvcs.constants import(COMMIT_FILE,COMPRESSION_FORMAT,META_ZIP,METAMODELS_FILE,NIL_PTR,STATE_ZIP,VER_SYMLINK,VERSION_SERVICE_INFO_FILE,VERSION_SPACE_ARCHIVE)
from localstack_ext.bootstrap.cpvcs.models import Commit,Revision,StateFileRef,Version
from localstack_ext.bootstrap.cpvcs.obj_storage import default_storage as object_storage
from localstack_ext.bootstrap.cpvcs.utils.common import config_context
from localstack_ext.bootstrap.cpvcs.utils.hash_utils import(compute_file_hash,compute_revision_hash,compute_version_archive_hash,random_hash)
from localstack_ext.bootstrap.cpvcs.utils.remote_utils import(_metadata_create_func,extract_meta_and_state_archives,merge_version_space,register_remote)
from localstack_ext.bootstrap.state_merge import merge_object_state
from localstack_ext.bootstrap.state_utils import(API_STATES_DIR,api_states_traverse,load_persisted_object,persist_object)
LOG=logging.getLogger(__name__)
def init(pod_name="My-Pod"):
 if config_context.pod_exists_locally(pod_name=pod_name):
  LOG.warning(f"Pod with name {pod_name} already exists locally")
  return
 config_context.set_pod_context(pod_name)
 def _create_internal_fs():
  mkdir(config_context.get_pod_root_dir())
  mkdir(config_context.get_ver_refs_path())
  mkdir(config_context.get_rev_refs_path())
  mkdir(config_context.get_ver_obj_store_path())
  mkdir(config_context.get_rev_obj_store_path())
  mkdir(config_context.get_delta_log_path())
 _create_internal_fs()
 r0_hash=random_hash()
 v0_hash=random_hash()
 r0=Revision(hash_ref=r0_hash,parent_ptr=NIL_PTR,creator=config_context.get_context_user(),rid=short_uid(),revision_number=0,state_files={})
 v0=Version(hash_ref=v0_hash,parent_ptr=NIL_PTR,creator=config_context.get_context_user(),comment="Init version",active_revision_ptr=r0_hash,outgoing_revision_ptrs={r0_hash},incoming_revision_ptr=FsTdk,state_files=FsTdi(),version_number=0)
 rev_key,ver_key=object_storage.upsert_objects(r0,v0)
 ver_symlink=config_context.create_version_symlink(VER_SYMLINK.format(ver_no=v0.version_number),ver_key)
 with FsTdW(config_context.get_head_path(),"w")as fp:
  fp.write(ver_symlink)
 with FsTdW(config_context.get_max_ver_path(),"w")as fp:
  fp.write(ver_symlink)
 with FsTdW(config_context.get_known_ver_path(),"w")as fp:
  fp.write(ver_symlink)
 config_context.update_ver_log(author=config_context.get_context_user(),ver_no=v0.version_number,rev_id=r0.rid,rev_no=r0.revision_number)
 LOG.debug(f"Successfully initated CPVCS for pod at {config_context.get_pod_root_dir()}")
def init_remote(version_space_archive:FsTdS,meta_archives:Dict[FsTdS,FsTdS],state_archives:Dict[FsTdS,FsTdS],remote_info:Dict[FsTdS,FsTdS],pod_name:FsTdS):
 config_context.set_pod_context(pod_name=pod_name)
 mkdir(config_context.get_pod_root_dir())
 with zipfile.ZipFile(version_space_archive)as version_space_zip:
  version_space_zip.extractall(config_context.get_pod_root_dir())
  LOG.debug("Successfully extracted version space zip")
 rm_rf(version_space_archive)
 extract_meta_and_state_archives(meta_archives=meta_archives,state_archives=state_archives)
 max_ver=_get_max_version()
 ver_symlink=config_context.create_version_symlink(name=VER_SYMLINK.format(ver_no=max_ver.version_number))
 with FsTdW(config_context.get_head_path(),"w")as fp:
  fp.write(ver_symlink)
 register_remote(remote_info=remote_info)
def merge_from_remote(version_space_archive:FsTdS,meta_archives:Dict[FsTdS,FsTdS],state_archives:Dict[FsTdS,FsTdS]):
 merge_version_space(version_space_archive)
 extract_meta_and_state_archives(meta_archives=meta_archives,state_archives=state_archives)
def is_remotely_managed()->FsTdO:
 return config_context.is_remotly_managed()
def set_pod_context(pod_name:FsTdS):
 config_context.set_pod_context(pod_name)
def rename_pod(new_pod_name:FsTdS)->FsTdO:
 if config_context.pod_exists_locally(new_pod_name):
  LOG.warning(f"{new_pod_name} already exists locally")
  return FsTdA
 config_context.rename_pod(new_pod_name)
 return FsTdq
def list_locally_available_pods(show_remote_or_local:FsTdO=FsTdq)->Set[FsTdS]:
 available_pods=os.listdir(config_context.cpvcs_root_dir)
 if not show_remote_or_local:
  return available_pods
 result=FsTdi()
 for available_pod in available_pods:
  pod_name=(f"remote/{available_pod}" if config_context.is_remotly_managed(available_pod)else f"local/{available_pod}")
  result.add(pod_name)
 return result
def create_state_file_from_fs(path:FsTdS,file_name:FsTdS,service:FsTdS,region:FsTdS)->FsTdS:
 file_path=os.path.join(path,file_name)
 key=compute_file_hash(file_path)
 rel_path=path.split(f"{API_STATES_DIR}/")[1]
 shutil.copy(file_path,os.path.join(config_context.get_obj_store_path(),key))
 state_file=StateFileRef(hash_ref=key,rel_path=rel_path,file_name=file_name,size=os.path.getsize(file_path),service=service,region=region)
 _add_state_file_to_expansion_point(state_file)
 return key
def _create_state_file_from_in_memory_blob(blob)->FsTdS:
 tmp_file_name=random_hash()
 tmp_dest=os.path.join(config_context.get_obj_store_path(),tmp_file_name)
 persist_object(blob,tmp_dest)
 key=compute_file_hash(tmp_dest)
 dest=os.path.join(config_context.get_obj_store_path(),key)
 os.rename(tmp_dest,dest)
 return key
def _get_state_file_path(key:FsTdS)->FsTdS:
 file_path=os.path.join(config_context.get_obj_store_path(),key)
 if os.path.isfile(file_path):
  return file_path
 LOG.warning(f"No state file with found with key: {key}")
def _add_state_file_to_expansion_point(state_file:StateFileRef):
 revision,_=_get_expansion_point_with_head()
 updated_state_files=FsTdi(FsTdn(lambda sf:not sf.congruent(state_file),revision.state_files))
 updated_state_files.add(state_file)
 revision.state_files=updated_state_files
 object_storage.upsert_objects(revision)
def list_state_files(key:FsTdS)->Optional[FsTdS]:
 cpvcs_obj=object_storage.get_revision_or_version_by_key(key)
 if cpvcs_obj:
  return cpvcs_obj.state_files_info()
 LOG.debug(f"No Version or Revision associated to {key}")
def get_version_info(version_no:FsTdx)->Dict[FsTdS,FsTdS]:
 archive_path=get_version_meta_archive(version_no)
 if not archive_path:
  LOG.warning(f"No Info found for version {version_no}")
  return
 return read_file_from_archive(archive_path,VERSION_SERVICE_INFO_FILE)
def read_file_from_archive(archive_path:FsTdS,file_name:FsTdS):
 try:
  with zipfile.ZipFile(archive_path)as archive:
   content=json.loads(archive.read(file_name))
   return json.dumps(content)
 except FsTdE as e:
  LOG.debug(f"Could not find {file_name} in archive {archive_path}: {e}")
def commit(message:FsTdS=FsTdk)->Revision:
 curr_expansion_point,head_version=_get_expansion_point_with_head()
 curr_expansion_point_hash=compute_revision_hash(curr_expansion_point)
 curr_expansion_point_parent_state_files=FsTdi()
 if curr_expansion_point.parent_ptr!=NIL_PTR:
  referenced_by_version=FsTdk
  curr_expansion_point_parent=object_storage.get_revision_by_key(curr_expansion_point.parent_ptr)
  curr_expansion_point_parent_state_files=curr_expansion_point_parent.state_files
  curr_expansion_point_parent.assoc_commit.head_ptr=curr_expansion_point_hash
  object_storage.upsert_objects(curr_expansion_point_parent)
 else:
  referenced_by_version=head_version.hash_ref
 object_storage.update_revision_key(curr_expansion_point.hash_ref,curr_expansion_point_hash,referenced_by_version)
 curr_expansion_point.hash_ref=curr_expansion_point_hash
 new_expansion_point=Revision(hash_ref=random_hash(),state_files={},parent_ptr=curr_expansion_point_hash,creator=curr_expansion_point.creator,rid=short_uid(),revision_number=curr_expansion_point.revision_number+1)
 delta_log_ptr=_create_delta_log(curr_expansion_point_parent_state_files,curr_expansion_point.state_files)
 assoc_commit=Commit(tail_ptr=curr_expansion_point.hash_ref,head_ptr=new_expansion_point.hash_ref,message=message,delta_log_ptr=delta_log_ptr)
 curr_expansion_point.assoc_commit=assoc_commit
 object_storage.upsert_objects(new_expansion_point,curr_expansion_point)
 config_context.update_ver_log(author=new_expansion_point.creator,ver_no=head_version.version_number,rev_id=new_expansion_point.rid,rev_no=new_expansion_point.revision_number)
 return curr_expansion_point
def zip_directories(zip_dest:FsTdS,directories:List[FsTdS])->FsTdS:
 for version_space_dir in directories:
  create_zip_file_python(source_path=version_space_dir,content_root=os.path.basename(version_space_dir),base_dir=version_space_dir,zip_file=zip_dest,mode="a")
 return zip_dest
def create_version_space_archive()->FsTdS:
 zip_dest=os.path.join(config_context.get_pod_root_dir(),VERSION_SPACE_ARCHIVE)
 rm_rf(zip_dest)
 result=zip_directories(zip_dest=zip_dest,directories=config_context.get_version_space_dir_paths())
 with zipfile.ZipFile(result,"a")as archive:
  for version_space_file in config_context.get_version_space_file_paths():
   archive.write(version_space_file,arcname=os.path.basename(version_space_file))
 return result
def get_head()->Version:
 return object_storage.get_version_by_key(config_context._get_head_key())
def _get_max_version()->Version:
 return object_storage.get_version_by_key(config_context.get_max_ver_key())
def get_max_version_no()->FsTdx:
 with FsTdW(config_context.get_max_ver_path())as fp:
  return FsTdx(os.path.basename(fp.readline()))
def _get_expansion_point_with_head()->Tuple[Revision,Version]:
 head_version=get_head()
 active_revision_root=object_storage.get_revision_by_key(head_version.active_revision_ptr)
 expansion_point=object_storage.get_terminal_revision(active_revision_root)
 return expansion_point,head_version
def _filter_special_cases(state_files:Set[StateFileRef])->Tuple[List[StateFileRef],List[StateFileRef],List[StateFileRef]]:
 regular_refs,s3_bucket_refs,sqs_queue_refs=[],[],[]
 for state_file in state_files:
  if state_file.service=="sqs":
   sqs_queue_refs.append(state_file)
  elif state_file.service=="s3":
   s3_bucket_refs.append(state_file)
  else:
   regular_refs.append(state_file)
 return regular_refs,s3_bucket_refs,sqs_queue_refs
def push_overwrite(version:FsTdx,comment:FsTdS)->FsTdO:
 expansion_point,_=_get_expansion_point_with_head()
 if version>get_max_version_no():
  LOG.debug("Attempted to overwrite a non existing version.. Aborting")
  return FsTdA
 version_node=get_version_by_number(version)
 _create_state_directory(version_number=version,state_file_refs=expansion_point.state_files)
 metamodels_file=f"metamodel_commit_{expansion_point.revision_number}.json"
 _create_metadata_archive(version_node=version_node,overwrite=FsTdq,metamodels_file=metamodels_file)
 version_node.comment=comment
 object_storage.upsert_objects(version_node)
 return FsTdq
def push(comment:FsTdS=FsTdk,three_way:FsTdO=FsTdA)->Version:
 expansion_point,head_version=_get_expansion_point_with_head()
 max_version=_get_max_version()
 new_active_revision=Revision(hash_ref=random_hash(),state_files=FsTdi(),parent_ptr=NIL_PTR,creator=expansion_point.creator,rid=short_uid(),revision_number=0)
 new_max_version_no=max_version.version_number+1
 if head_version.version_number!=max_version.version_number:
  _merge_expansion_point_with_max(three_way=three_way)
 else:
  _create_state_directory(new_max_version_no,state_file_refs=expansion_point.state_files)
 new_version=Version(hash_ref=compute_version_archive_hash(new_max_version_no,get_version_state_archive(new_max_version_no)),state_files=FsTdi(),parent_ptr=max_version.hash_ref,creator=expansion_point.creator,comment=comment,active_revision_ptr=new_active_revision.hash_ref,outgoing_revision_ptrs={new_active_revision.hash_ref},incoming_revision_ptr=expansion_point.hash_ref,version_number=new_max_version_no)
 if expansion_point.parent_ptr!=NIL_PTR:
  expansion_point_parent=object_storage.get_revision_by_key(expansion_point.parent_ptr)
  state_from=expansion_point_parent.state_files
 else:
  state_from=FsTdi()
 delta_log_ptr=_create_delta_log(state_from,new_version.state_files)
 expansion_point_commit=Commit(tail_ptr=expansion_point.hash_ref,head_ptr=new_version.hash_ref,message="Finalizing commit",delta_log_ptr=delta_log_ptr)
 expansion_point.state_files=new_version.state_files
 expansion_point.assoc_commit=expansion_point_commit
 head_version.active_revision_ptr=NIL_PTR
 object_storage.upsert_objects(head_version,expansion_point,new_active_revision,new_version)
 _update_head(new_version.version_number,new_version.hash_ref)
 _update_max_ver(new_version.version_number,new_version.hash_ref)
 _add_known_ver(new_version.version_number,new_version.hash_ref)
 _create_metadata_archive(new_version)
 config_context.update_ver_log(author=expansion_point.creator,ver_no=new_version.version_number,rev_id=new_active_revision.rid,rev_no=new_active_revision.revision_number)
 return new_version
def get_version_meta_archive(version_no:FsTdx)->FsTdS:
 version_meta_path=config_context.get_version_meta_archive_path(version_no)
 if os.path.isfile(version_meta_path):
  return version_meta_path
def get_version_state_archive(version_no:FsTdx)->FsTdS:
 version_state_path=config_context.get_version_state_archive_path(version_no)
 if os.path.isfile(version_state_path):
  return version_state_path
def get_version_metamodel(version_no:FsTdx)->FsTdS:
 meta_archive=get_version_meta_archive(version_no)
 if meta_archive:
  return read_file_from_archive(meta_archive,METAMODELS_FILE)
def _create_metadata_archive(version_node:Version,delete_reference:FsTdO=FsTdA,overwrite:FsTdO=FsTdA,metamodels_file:FsTdS=FsTdk):
 revision_node=object_storage.get_revision_or_version_by_key(version_node.active_revision_ptr if overwrite else version_node.incoming_revision_ptr)
 metadata_dir=os.path.join(config_context.get_pod_root_dir(),META_ZIP.format(version_no=version_node.version_number))
 mkdir(metadata_dir)
 if metamodels_file:
  archive_path=f"{metadata_dir}.{COMPRESSION_FORMAT}"
  if os.path.isfile(archive_path):
   with zipfile.ZipFile(archive_path)as archive:
    archive.extractall(metadata_dir)
 while revision_node:
  assoc_commit=revision_node.assoc_commit
  if not assoc_commit:
   break
  delta_ptr=assoc_commit.delta_log_ptr
  if delta_ptr:
   src=object_storage.get_delta_file_by_key(delta_ptr)
   if not src:
    continue
   dst_name=COMMIT_FILE.format(commit_no=revision_node.revision_number+1)
   dst=os.path.join(config_context.get_pod_root_dir(),metadata_dir,dst_name)
   shutil.copy(src,dst)
   if delete_reference:
    os.remove(src)
  next_revision=assoc_commit.head_ptr if overwrite else revision_node.parent_ptr
  revision_node=object_storage.get_revision_by_key(next_revision)
 metamodels={}
 service_info={}
 tmp_states_dir=new_tmp_dir()
 with zipfile.ZipFile(get_version_state_archive(version_node.version_number))as archive:
  archive.extractall(tmp_states_dir)
 tmp_states_dir_api_states=os.path.join(tmp_states_dir,API_STATES_DIR)
 api_states_traverse(api_states_path=tmp_states_dir_api_states,side_effect=_metadata_create_func,mutables=[metamodels,service_info])
 if not metamodels_file:
  metamodels_file=METAMODELS_FILE
 metamodels_dest=os.path.join(metadata_dir,metamodels_file)
 version_service_info=os.path.join(metadata_dir,VERSION_SERVICE_INFO_FILE)
 with FsTdW(metamodels_dest,"w")as fp:
  metamodels=json_safe(metamodels)
  json.dump(metamodels,fp,indent=4)
 with FsTdW(version_service_info,"w")as fp:
  json.dump(service_info,fp,indent=4)
 shutil.make_archive(metadata_dir,COMPRESSION_FORMAT,root_dir=metadata_dir)
 rm_rf(metadata_dir)
 rm_rf(tmp_states_dir)
def _create_state_directory(version_number:FsTdx,state_file_refs:Set[StateFileRef],delete_files=FsTdA,archive=FsTdq):
 version_state_dir=os.path.join(config_context.get_pod_root_dir(),STATE_ZIP.format(version_no=version_number))
 for state_file in state_file_refs:
  try:
   dst_path=os.path.join(version_state_dir,API_STATES_DIR,state_file.rel_path)
   mkdir(dst_path)
   src=object_storage.get_state_file_location_by_key(state_file.hash_ref)
   dst=os.path.join(dst_path,state_file.file_name)
   shutil.copy(src,dst)
   if delete_files:
    os.remove(src)
  except FsTdE as e:
   LOG.warning(f"Failed to locate state file with rel path: {state_file.rel_path}: {e}")
 mkdir(os.path.join(version_state_dir,"kinesis"))
 mkdir(os.path.join(version_state_dir,"dynamodb"))
 if archive:
  shutil.make_archive(version_state_dir,COMPRESSION_FORMAT,root_dir=version_state_dir)
  rm_rf(version_state_dir)
  return f"{version_state_dir}.{COMPRESSION_FORMAT}"
 return version_state_dir
def set_active_version(version_no:FsTdx,commit_before=FsTdA)->FsTdO:
 known_versions=load_version_references()
 for known_version_no,known_version_key in known_versions:
  if known_version_no==version_no:
   if commit_before:
    commit()
   _set_active_version(known_version_key)
   return FsTdq
 LOG.info(f"Version with number {version_no} not found")
 return FsTdA
def _set_active_version(key:FsTdS):
 current_head=get_head()
 if current_head.hash_ref!=key and object_storage.version_exists(key):
  requested_version=object_storage.get_version_by_key(key)
  _update_head(requested_version.version_number,key)
  if requested_version.active_revision_ptr==NIL_PTR:
   new_path_root=Revision(hash_ref=random_hash(),state_files=FsTdi(),parent_ptr=NIL_PTR,creator=config_context.get_context_user(),rid=short_uid(),revision_number=0)
   requested_version.active_revision_ptr=new_path_root.hash_ref
   requested_version.outgoing_revision_ptrs.add(new_path_root.hash_ref)
   object_storage.upsert_objects(new_path_root,requested_version)
def get_version_by_number(version_no:FsTdx)->Version:
 versions=load_version_references()
 version_ref=FsTdB((version[1]for version in versions if version[0]==version_no),FsTdk)
 if not version_ref:
  LOG.warning(f"Could not find version number {version_no}")
  return
 return object_storage.get_version_by_key(version_ref)
def load_version_references()->List[Tuple[FsTdx,FsTdS]]:
 result={}
 with FsTdW(config_context.get_known_ver_path(),"r")as vp:
  symlinks=vp.readlines()
  for symlink in symlinks:
   symlink=config_context.get_pod_absolute_path(symlink.rstrip())
   with FsTdW(symlink,"r")as sp:
    result[FsTdx(os.path.basename(symlink))]=sp.readline()
 return FsTdj(result.items(),key=lambda x:x[0],reverse=FsTdq)
def list_versions()->List[FsTdS]:
 version_references=load_version_references()
 result=[object_storage.get_version_by_key(version_key).info_str()for _,version_key in version_references]
 return result
def list_version_commits(version_no:FsTdx)->List[FsTdS]:
 if version_no==-1:
  version=_get_max_version()
 else:
  version=get_version_by_number(version_no)
 if not version:
  return[]
 result=[]
 revision=object_storage.get_revision_by_key(version.incoming_revision_ptr)
 while revision:
  assoc_commit=revision.assoc_commit
  revision_no=revision.revision_number
  if revision_no!=0:
   from_node=f"Revision-{revision_no - 1}"
  elif version_no!=0:
   from_node=f"Version-{version_no}"
  else:
   from_node="Empty state"
  to_node=f"Revision-{revision_no}"
  result.append(assoc_commit.info_str(from_node=from_node,to_node=to_node))
  revision=object_storage.get_revision_by_key(revision.parent_ptr)
 return result
def get_commit_diff(version_no:FsTdx,commit_no:FsTdx)->FsTdS:
 archive_path=get_version_meta_archive(version_no)
 if not archive_path:
  LOG.warning(f"No metadata found for version {version_no}")
  return
 file_name=f"{COMMIT_FILE.format(commit_no=commit_no)}"
 return read_file_from_archive(archive_path=archive_path,file_name=file_name)
def _update_head(new_head_ver_no,new_head_key)->FsTdS:
 with FsTdW(config_context.get_head_path(),"w")as fp:
  ver_symlink=config_context.create_version_symlink(VER_SYMLINK.format(ver_no=new_head_ver_no),new_head_key)
  fp.write(ver_symlink)
  return ver_symlink
def _update_max_ver(new_max_ver_no,new_max_ver_key)->FsTdS:
 with FsTdW(config_context.get_max_ver_path(),"w")as fp:
  max_ver_symlink=config_context.create_version_symlink(VER_SYMLINK.format(ver_no=new_max_ver_no),new_max_ver_key)
  fp.write(max_ver_symlink)
  return max_ver_symlink
def _add_known_ver(new_ver_no,new_ver_key)->FsTdS:
 with FsTdW(config_context.get_known_ver_path(),"a")as fp:
  new_ver_symlink=config_context.create_version_symlink(VER_SYMLINK.format(ver_no=new_ver_no),new_ver_key)
  fp.write(f"\n{new_ver_symlink}")
  return new_ver_symlink
def _is_special_case(obj)->FsTdO:
 return FsTdy(obj,(Queue,FakeBucket))
def _merge_three_way_dir_func(**kwargs):
 dir_name=kwargs.get("dir_name")
 fname=kwargs.get("fname")
 region=kwargs.get("region")
 service_name=kwargs.get("service_name")
 mutables=kwargs.get("mutables")
 other=mutables[0]
 ancestor=mutables[1]
 src_state_file_path=os.path.join(dir_name,fname)
 ancestor_state_dir=os.path.join(ancestor,service_name,region)
 ancestor_state_file_path=os.path.join(ancestor_state_dir,fname)
 dst_state_dir=os.path.join(other,service_name,region)
 dst_state_file_path=os.path.join(dst_state_dir,fname)
 src_state=load_persisted_object(src_state_file_path)
 ancestor_state=load_persisted_object(ancestor_state_file_path)
 special_case=_is_special_case(src_state)
 if os.path.isfile(dst_state_file_path):
  if not special_case:
   dst_state=load_persisted_object(dst_state_file_path)
   merge_object_state(dst_state,src_state,ancestor_state)
   persist_object(dst_state,dst_state_file_path)
 else:
  mkdir(dst_state_dir)
  persist_object(src_state,dst_state_file_path)
def _merge_two_state_dir_func(**kwargs):
 dir_name=kwargs.get("dir_name")
 fname=kwargs.get("fname")
 region=kwargs.get("region")
 service_name=kwargs.get("service_name")
 mutables=kwargs.get("mutables")
 other=mutables[0]
 src_state_file_path=os.path.join(dir_name,fname)
 dst_state_dir=os.path.join(other,service_name,region)
 dst_state_file_path=os.path.join(dst_state_dir,fname)
 src_state=load_persisted_object(src_state_file_path)
 special_case=_is_special_case(src_state)
 if os.path.isfile(dst_state_file_path):
  if not special_case:
   dst_state=load_persisted_object(dst_state_file_path)
   merge_object_state(dst_state,src_state)
   persist_object(dst_state,dst_state_file_path)
 else:
  mkdir(dst_state_dir)
  persist_object(src_state,dst_state_file_path)
def _merge_expansion_point_with_max(three_way=FsTdA):
 expansion_point,head=_get_expansion_point_with_head()
 curr_max_version_no=get_max_version_no()
 new_version_no=curr_max_version_no+1
 curr_max_version_state_archive=get_version_state_archive(curr_max_version_no)
 common_ancestor_state_archive=FsTdk
 tmp_common_ancestor_state_dir=FsTdk
 if head.version_number>1 and three_way:
  common_ancestor_state_archive=get_version_state_archive(head.version_number-1)
  tmp_common_ancestor_state_dir=new_tmp_dir()
  with zipfile.ZipFile(common_ancestor_state_archive)as archive:
   archive.extractall(tmp_common_ancestor_state_dir)
 tmp_max_version_state_dir=new_tmp_dir()
 with zipfile.ZipFile(curr_max_version_state_archive)as archive:
  archive.extractall(tmp_max_version_state_dir)
 new_max_version_states_dir=_create_state_directory(version_number=new_version_no,state_file_refs=expansion_point.state_files,archive=FsTdA)
 new_max_version_states_dir_api_states=os.path.join(new_max_version_states_dir,API_STATES_DIR)
 tmp_max_version_state_dir_api_states=os.path.join(tmp_max_version_state_dir,API_STATES_DIR)
 if not common_ancestor_state_archive:
  api_states_traverse(api_states_path=tmp_max_version_state_dir_api_states,side_effect=_merge_two_state_dir_func,mutables=[new_max_version_states_dir_api_states])
 else:
  LOG.debug(f"Starting three way merge between version {head.version_number} and {curr_max_version_no} with common ancestor {head.version_number - 1}")
  common_ancestor_version_state_dir_api_states=os.path.join(tmp_common_ancestor_state_dir,API_STATES_DIR)
  api_states_traverse(api_states_path=tmp_max_version_state_dir_api_states,side_effect=_merge_three_way_dir_func,mutables=[new_max_version_states_dir_api_states,common_ancestor_version_state_dir_api_states])
 shutil.make_archive(new_max_version_states_dir,COMPRESSION_FORMAT,root_dir=new_max_version_states_dir)
 rm_rf(new_max_version_states_dir)
 rm_rf(tmp_max_version_state_dir)
 rm_rf(tmp_common_ancestor_state_dir)
def _infer_backend_init(clazz,sf):
 if FsTdy(clazz,FsTdC):
  backend={}
 else:
  constructor=FsTdo(clazz,"__init__",FsTdk)
  sig_args=inspect.getfullargspec(constructor)
  if "region" in sig_args.args:
   backend=clazz(region=sf.region)
  elif "region_name" in sig_args.args:
   backend=clazz(region_name=sf.region)
  else:
   backend=clazz()
 return backend
def _create_delta_log(state_from:Set[StateFileRef],state_to:Set[StateFileRef])->FsTdS:
 state_files_from=_filter_special_cases(state_from)
 state_files_from_regular=state_files_from[0]
 state_files_from_s3=state_files_from[1]
 state_files_from_sqs=state_files_from[2]
 state_files_to=_filter_special_cases(state_to)
 state_files_to_regular=state_files_to[0]
 state_files_to_s3=state_files_to[1]
 state_files_to_sqs=state_files_to[2]
 def _create_sf_lookup(state_files:Set[StateFileRef])->Dict[FsTdS,StateFileRef]:
  return{os.path.join(sf.rel_path,sf.file_name):sf for sf in state_files}
 result={}
 state_files_to_lookup=_create_sf_lookup(state_files_to_regular)
 for state_file_from in state_files_from_regular:
  result_region=result.setdefault(state_file_from.region,{})
  result_service=result_region.setdefault(state_file_from.service,{})
  result_file=result_service.setdefault(state_file_from.file_name,{})
  if state_file_from.any_congruence(state_files_to_regular):
   key=os.path.join(state_file_from.rel_path,state_file_from.file_name)
   state_file_to=state_files_to_lookup.pop(key)
   diff_json=_create_diff_json(state_file_from,state_file_to)
  else:
   diff_json=_create_diff_json(state_file_from,FsTdk)
  result_file["diff"]=diff_json
 for state_files_to in state_files_to_lookup.values():
  result_region=result.setdefault(state_files_to.region,{})
  result_service=result_region.setdefault(state_files_to.service,{})
  result_file=result_service.setdefault(state_files_to.file_name,{})
  diff_json=_create_diff_json(FsTdk,state_files_to)
  result_file["diff"]=diff_json
 def _handle_special_case_containers(service,sf_from,sf_to):
  region_containers_from={}
  region_containers_to={}
  for container_file_from in sf_from:
   region=container_file_from.region
   region_container=region_containers_from.setdefault(region,{})
   container=load_persisted_object(object_storage.get_state_file_location_by_key(container_file_from.hash_ref))
   if container:
    region_container[container.name]=container
  for container_file_to in sf_to:
   region=container_file_to.region
   region_container=region_containers_to.setdefault(region,{})
   container=load_persisted_object(object_storage.get_state_file_location_by_key(container_file_to.hash_ref))
   if container:
    region_container[container.name]=container
  for region,region_queues_from in region_containers_from.items():
   region_queues_to=region_containers_to.pop(region,{})
   region_diff=DeepDiff(region_queues_from,region_queues_to)
   container_result_region=result.setdefault(region,{})
   container_result_service=container_result_region.setdefault(service,{})
   container_result_service["diff"]=region_diff.to_json()
  for region,region_queues_to in region_containers_to.items():
   region_diff=DeepDiff({},region_queues_to)
   container_result_region=result.setdefault(region,{})
   container_result_service=container_result_region.setdefault(service,{})
   container_result_service["diff"]=region_diff.to_json()
 _handle_special_case_containers("sqs",state_files_from_sqs,state_files_to_sqs)
 _handle_special_case_containers("s3",state_files_from_s3,state_files_to_s3)
 tmp_dest=os.path.join(config_context.get_delta_log_path(),random_hash())
 with FsTdW(tmp_dest,"w")as fp:
  json.dump(result,fp,indent=4)
 key=compute_file_hash(tmp_dest)
 dest=os.path.join(config_context.get_delta_log_path(),key)
 os.rename(tmp_dest,dest)
 return key
def _create_diff_json(sf1:StateFileRef,sf2:StateFileRef):
 if sf1 and sf2:
  if sf1.hash_ref==sf2.hash_ref:
   return "No Changes"
  else:
   backend1=load_persisted_object(config_context.get_obj_file_path(sf1.hash_ref))
   backend2=load_persisted_object(config_context.get_obj_file_path(sf2.hash_ref))
   diff=DeepDiff(backend1,backend2)
   return diff.to_json()
 if not sf1:
  backend2=load_persisted_object(config_context.get_obj_file_path(sf2.hash_ref))
  clazz=FsTdL(backend2)
  backend1=_infer_backend_init(clazz,sf2)
 else:
  backend1=load_persisted_object(config_context.get_obj_file_path(sf1.hash_ref))
  clazz=FsTdL(backend1)
  backend2=_infer_backend_init(clazz,sf1)
 diff=DeepDiff(backend1,backend2)
 return diff.to_json()
# Created by pyminifier (https://github.com/liftoff/pyminifier)



from localstack.services.cloudformation.service_models import GenericBaseModel
YnfSv=staticmethod
YnfSy=classmethod
from localstack.utils.aws import aws_stack
class IoTCertificate(GenericBaseModel):
 @YnfSv
 def cloudformation_type():
  return "AWS::IoT::Certificate"
 def fetch_state(self,stack_name,resources):
  return self.state.get("certificateId")
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("certificateId")
 @YnfSy
 def get_deploy_templates(cls):
  def create_certificate_from_csr(resource_id,resources,*args,**kwargs):
   client=aws_stack.connect_to_service("iot")
   resource=cls(resources[resource_id])
   sign_req=resource.props.get("CertificateSigningRequest")
   response=client.create_certificate_from_csr(certificateSigningRequest=sign_req)
   resource.state["certificateId"]=response.get("certificateId")
   return response
  return{"create":{"function":create_certificate_from_csr}}
# Created by pyminifier (https://github.com/liftoff/pyminifier)



from localstack.services.cloudformation.service_models import GenericBaseModel
IJcoC=staticmethod
IJcor=None
from localstack.utils.aws import aws_stack
class ApiGatewayAuthorizer(GenericBaseModel):
 @IJcoC
 def cloudformation_type():
  return "AWS::ApiGateway::Authorizer"
 def fetch_state(self,stack_name,resources):
  props=self.props
  client=aws_stack.connect_to_service("apigateway")
  api_id=self.resolve_refs_recursively(stack_name,props["RestApiId"],resources)
  auth_uri=self.resolve_refs_recursively(stack_name,props.get("AuthorizerUri"),resources)
  authorizers=client.get_authorizers(restApiId=api_id,limit=200)["items"]
  result=[a for a in authorizers if a["type"]==props.get("Type")and a.get("authorizerUri")==auth_uri]
  return(result or[IJcor])[0]
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("id")
# Created by pyminifier (https://github.com/liftoff/pyminifier)



import logging
pIWcN=bool
pIWck=hasattr
pIWcX=set
pIWcs=True
pIWcQ=False
pIWcU=isinstance
pIWch=dict
pIWcO=getattr
pIWcg=None
pIWcv=str
pIWcD=Exception
pIWco=open
import os
from typing import Any,Callable,List,OrderedDict,Set,Tuple
import dill
from localstack.utils.common import ObjectIdHashComparator
API_STATES_DIR="api_states"
LOG=logging.getLogger(__name__)
def check_already_visited(obj,visited:Set)->Tuple[pIWcN,Set]:
 if pIWck(obj,"__dict__"):
  visited=visited or pIWcX()
  wrapper=ObjectIdHashComparator(obj)
  if wrapper in visited:
   return pIWcs,visited
  visited.add(wrapper)
 return pIWcQ,visited
def get_object_dict(obj):
 if pIWcU(obj,pIWch):
  return obj
 obj_dict=pIWcO(obj,"__dict__",pIWcg)
 return obj_dict
def is_composite_type(obj):
 return pIWcU(obj,(pIWch,OrderedDict))or pIWck(obj,"__dict__")
def api_states_traverse(api_states_path:pIWcv,side_effect:Callable[...,pIWcg],mutables:List[Any]):
 for dir_name,_,file_list in os.walk(api_states_path):
  for file_name in file_list:
   try:
    subdirs=os.path.normpath(dir_name).split(os.sep)
    region=subdirs[-1]
    service_name=subdirs[-2]
    side_effect(dir_name=dir_name,fname=file_name,region=region,service_name=service_name,mutables=mutables)
   except pIWcD as e:
    LOG.warning(f"Failed to apply {side_effect.__name__} for {file_name} in dir {dir_name}: {e}")
    continue
def load_persisted_object(state_file):
 if not os.path.isfile(state_file):
  return
 import dill
 with pIWco(state_file,"rb")as f:
  try:
   content=f.read()
   result=dill.loads(content)
   return result
  except pIWcD as e:
   LOG.debug("Unable to read pickled persistence file %s: %s"%(state_file,e))
def persist_object(obj,state_file):
 with pIWco(state_file,"wb")as f:
  result=f.write(dill.dumps(obj))
  return result
# Created by pyminifier (https://github.com/liftoff/pyminifier)



from localstack.services.cloudformation.deployment_utils import(lambda_keys_to_lower,params_list_to_dict)
xPNCt=staticmethod
xPNCV=None
xPNCT=super
xPNCz=classmethod
from localstack.services.cloudformation.service_models import REF_ID_ATTRS,GenericBaseModel
from localstack.utils.aws import aws_stack
from localstack_ext.utils.aws import aws_utils
class AmplifyBranch(GenericBaseModel):
 @xPNCt
 def cloudformation_type():
  return "AWS::Amplify::Branch"
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("amplify")
  props=self.props
  app_id=self.resolve_refs_recursively(stack_name,props["AppId"],resources)
  branch_name=self.resolve_refs_recursively(stack_name,props["BranchName"],resources)
  branches=client.list_branches(appId=app_id)["branches"]
  branch=[b for b in branches if b["branchName"]==branch_name]
  return(branch or[xPNCV])[0]
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("BranchName")
 @xPNCt
 def get_deploy_templates():
  def get_amplify_branch_params(params,**kwargs):
   params=lambda_keys_to_lower()(params,**kwargs)
   auth_config=params.pop("basicAuthConfig",{})
   if auth_config:
    params["enableBasicAuth"]=auth_config["EnableBasicAuth"]
    params["basicAuthCredentials"]="%s:%s"%(auth_config["Username"],auth_config["Password"])
   params["tags"]=params_list_to_dict("tags",key_attr_name="key",value_attr_name="value")(params,**kwargs)
   return params
  return{"create":{"function":"create_branch","parameters":get_amplify_branch_params}}
class AmplifyApp(GenericBaseModel):
 @xPNCt
 def cloudformation_type():
  return "AWS::Amplify::App"
 def get_cfn_attribute(self,attribute_name):
  if attribute_name=="DefaultDomain":
   return "default.example.com"
  if attribute_name in REF_ID_ATTRS:
   return self.get_physical_resource_id(attribute_name)
  if attribute_name in["Arn"]:
   return self.props.get("appArn")
  return xPNCT(AmplifyApp,self).get_cfn_attribute(attribute_name)
 def get_physical_resource_id(self,attribute,**kwargs):
  props=self.props
  app_id=props.get("appId")
  if not app_id:
   return
  if attribute in REF_ID_ATTRS:
   return app_id
  return props.get("appArn")or aws_utils.amplify_app_arn(app_id)
 @xPNCz
 def fetch_details(self,app_name):
  client=aws_stack.connect_to_service("amplify")
  apps=client.list_apps()["apps"]
  apps=[a for a in apps if a["name"]==app_name]
  return(apps or[xPNCV])[0]
 def fetch_state(self,stack_name,resources):
  app_name=self.resolve_refs_recursively(stack_name,self.props["Name"],resources)
  return self.fetch_details(app_name)
 @xPNCt
 def get_deploy_templates():
  def get_amplify_app_params(params,**kwargs):
   if "IAMServiceRole" in params:
    params["iamServiceRoleArn"]=params.pop("IAMServiceRole")
   params=lambda_keys_to_lower()(params,**kwargs)
   params["tags"]=params_list_to_dict("tags",key_attr_name="key",value_attr_name="value")(params,**kwargs)
   return params
  return{"create":{"function":"create_app","parameters":get_amplify_app_params}}
# Created by pyminifier (https://github.com/liftoff/pyminifier)



from localstack.services.cloudformation.service_models import REF_ID_ATTRS,GenericBaseModel
TXvRQ=staticmethod
TXvRL=None
from localstack.utils.aws import aws_stack
class CloudTrail(GenericBaseModel):
 @TXvRQ
 def cloudformation_type():
  return "AWS::CloudTrail::Trail"
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("cloudtrail")
  result=[t for t in client.list_trails()["Trails"]if t["Name"]==self.props["TrailName"]]
  return(result or[TXvRL])[0]
 def get_physical_resource_id(self,attribute,**kwargs):
  if attribute in REF_ID_ATTRS:
   return self.props["TrailName"]
 @TXvRQ
 def get_deploy_templates():
  def put_event_selectors(resource_id,resources,*args,**kwargs):
   resource=resources[resource_id]
   props=resource.get("Properties",{})
   selectors=props.get("EventSelectors",[])
   if selectors:
    cloudtrail=aws_stack.connect_to_service("cloudtrail")
    cloudtrail.put_event_selectors(TrailName=props["TrailName"],EventSelectors=selectors)
   result={}
   return result
  return{"create":[{"function":"create_trail","parameters":{"Name":"TrailName","S3BucketName":"S3BucketName","S3KeyPrefix":"S3KeyPrefix","SnsTopicName":"SnsTopicName","IncludeGlobalServiceEvents":"IncludeGlobalServiceEvents","IsMultiRegionTrail":"IsMultiRegionTrail","EnableLogFileValidation":"EnableLogFileValidation","CloudWatchLogsLogGroupArn":"CloudWatchLogsLogGroupArn","CloudWatchLogsRoleArn":"CloudWatchLogsRoleArn","KmsKeyId":"KMSKeyId"}},{"function":put_event_selectors}],"delete":{"function":"delete_trail","parameters":{"Name":"TrailName"}}}
# Created by pyminifier (https://github.com/liftoff/pyminifier)



import logging
XtIJi=None
XtIJB=str
XtIJP=True
XtIJd=int
XtIJG=open
XtIJO=bool
XtIJg=classmethod
import os
from typing import List
from localstack import config as localstack_config
from localstack_ext.bootstrap.auth import get_auth_cache
from localstack_ext.bootstrap.cpvcs.constants import(COMPRESSION_FORMAT,CPVCS_DIR,DEFAULT_POD_DIR,DELTA_LOG_DIR,HEAD_FILE,KNOWN_VER_FILE,MAX_VER_FILE,META_ZIP,OBJ_STORE_DIR,REFS_DIR,REMOTE_FILE,REV_SUB_DIR,STATE_ZIP,VER_LOG_FILE,VER_LOG_STRUCTURE,VER_SUB_DIR,VERSION_SPACE_DIRS,VERSION_SPACE_FILES)
LOG=logging.getLogger(__name__)
class CPVCSConfigContext:
 default_instance=XtIJi
 def __init__(self,pod_root_dir:XtIJB):
  self.cpvcs_root_dir=pod_root_dir
  self.pod_root_dir=pod_root_dir
  self.user=XtIJi
 def get_pod_context(self)->XtIJB:
  return os.path.basename(self.pod_root_dir)
 def get_context_user(self)->XtIJB:
  return self.user
 def get_pod_root_dir(self)->XtIJB:
  return self.pod_root_dir
 def get_head_path(self)->XtIJB:
  return os.path.join(self.pod_root_dir,HEAD_FILE)
 def get_max_ver_path(self)->XtIJB:
  return os.path.join(self.pod_root_dir,MAX_VER_FILE)
 def get_known_ver_path(self)->XtIJB:
  return os.path.join(self.pod_root_dir,KNOWN_VER_FILE)
 def get_ver_log_path(self)->XtIJB:
  return os.path.join(self.pod_root_dir,VER_LOG_FILE)
 def get_obj_store_path(self)->XtIJB:
  return os.path.join(self.pod_root_dir,OBJ_STORE_DIR)
 def get_rev_obj_store_path(self)->XtIJB:
  return os.path.join(self.get_obj_store_path(),REV_SUB_DIR)
 def get_ver_obj_store_path(self)->XtIJB:
  return os.path.join(self.get_obj_store_path(),VER_SUB_DIR)
 def get_ver_refs_path(self)->XtIJB:
  return os.path.join(self.pod_root_dir,REFS_DIR,VER_SUB_DIR)
 def get_rev_refs_path(self)->XtIJB:
  return os.path.join(self.pod_root_dir,REFS_DIR,REV_SUB_DIR)
 def get_version_ref_file_path(self,version_ref:XtIJB)->XtIJB:
  return os.path.join(self.get_ver_refs_path(),version_ref)
 def get_delta_log_path(self)->XtIJB:
  return os.path.join(self.pod_root_dir,self.get_obj_store_path(),DELTA_LOG_DIR)
 def get_version_meta_archive_path(self,version_no,with_format:XtIJO=XtIJP)->XtIJB:
  version_meta_path=os.path.join(self.get_pod_root_dir(),META_ZIP.format(version_no=version_no))
  if not with_format:
   return version_meta_path
  return f"{version_meta_path}.{COMPRESSION_FORMAT}"
 def get_version_state_archive_path(self,version_no,with_format:XtIJO=XtIJP)->XtIJB:
  version_state_path=os.path.join(self.get_pod_root_dir(),STATE_ZIP.format(version_no=version_no))
  if not with_format:
   return version_state_path
  return f"{version_state_path}.{COMPRESSION_FORMAT}"
 def update_ver_log(self,author:XtIJB,ver_no:XtIJd,rev_id:XtIJB,rev_no:XtIJd):
  with XtIJG(self.get_ver_log_path(),"a")as fp:
   fp.write(f"{VER_LOG_STRUCTURE.format(author=author, ver_no=ver_no, rev_rid_no=f'{rev_id}_{rev_no}')}\n")
 def create_version_symlink(self,name:XtIJB,key:XtIJB=XtIJi)->XtIJB:
  return self._create_symlink(name,key,self.get_ver_refs_path())
 def create_revision_symlink(self,name:XtIJB,key:XtIJB=XtIJi)->XtIJB:
  return self._create_symlink(name,key,self.get_rev_refs_path())
 def is_initialized(self)->XtIJO:
  return self.pod_root_dir and os.path.isdir(self.pod_root_dir)
 def _create_symlink(self,name:XtIJB,key:XtIJB,path:XtIJB)->XtIJB:
  rel_path=os.path.relpath(path,start=self.get_pod_root_dir())
  rel_symlink=os.path.join(rel_path,name)
  if key:
   symlink=os.path.join(path,name)
   with XtIJG(symlink,"w")as fp:
    fp.write(key)
  return rel_symlink
 def _get_head_key(self)->XtIJB:
  return self._get_key(self.get_head_path())
 def get_max_ver_key(self)->XtIJB:
  return self._get_key(self.get_max_ver_path())
 def _get_key(self,path:XtIJB)->XtIJB:
  with XtIJG(path,"r")as fp:
   rel_key_path=fp.readline().strip()
  key_path=self.get_pod_absolute_path(rel_key_path)
  with XtIJG(key_path,"r")as fp:
   key=fp.readline()
   return key
 def get_pod_absolute_path(self,rel_path):
  return os.path.join(self.get_pod_root_dir(),rel_path)
 def get_obj_file_path(self,key:XtIJB)->XtIJB:
  return os.path.join(self.get_obj_store_path(),key)
 def get_remote_info_path(self)->XtIJB:
  return os.path.join(self.pod_root_dir,REMOTE_FILE)
 def is_remotly_managed(self,pod_name:XtIJB=XtIJi)->XtIJO:
  if pod_name:
   return os.path.isfile(os.path.join(self.cpvcs_root_dir,pod_name,REMOTE_FILE))
  else:
   return os.path.isfile(self.get_remote_info_path())
 def set_pod_context(self,pod_name:XtIJB):
  cache=get_auth_cache()
  user=cache.get("username","unknown")
  self.pod_root_dir=os.path.join(self.cpvcs_root_dir,pod_name)
  self.user=user
 def pod_exists_locally(self,pod_name:XtIJB)->XtIJO:
  return os.path.isdir(os.path.join(self.cpvcs_root_dir,pod_name))
 def rename_pod(self,new_pod_name:XtIJB):
  curr_name=self.get_pod_root_dir()
  new_name=os.path.join(self.cpvcs_root_dir,new_pod_name)
  os.rename(curr_name,new_name)
  self.set_pod_context(new_name)
 def get_pod_name(self)->XtIJB:
  return os.path.basename(self.get_pod_root_dir())
 def get_version_space_dir_paths(self)->List[XtIJB]:
  return[os.path.join(self.get_pod_root_dir(),directory)for directory in VERSION_SPACE_DIRS]
 def get_version_space_file_paths(self)->List[XtIJB]:
  return[os.path.join(self.get_pod_root_dir(),filename)for filename in VERSION_SPACE_FILES]
 @XtIJg
 def get(cls):
  if not cls.default_instance:
   pod_root_dir=os.environ.get("POD_DIR")
   if not pod_root_dir:
    pod_root_dir=os.path.join(localstack_config.dirs.tmp,DEFAULT_POD_DIR)
   pod_root_dir=os.path.join(pod_root_dir,CPVCS_DIR)
   cls.default_instance=CPVCSConfigContext(pod_root_dir)
  return cls.default_instance
config_context=CPVCSConfigContext.get()
# Created by pyminifier (https://github.com/liftoff/pyminifier)



from localstack import config
hJBnk=staticmethod
hJBnv=None
hJBnW=classmethod
hJBnE=str
hJBnQ=False
hJBnP=True
hJBno=super
from localstack.services.cloudformation.service_models import GenericBaseModel
from localstack.utils.aws import aws_stack
from localstack.utils.common import clone,items_equivalent
from localstack_ext import config as ext_config
from localstack_ext.services.cloudformation.service_models import LOG
class ELBV2TargetGroup(GenericBaseModel):
 @hJBnk
 def cloudformation_type():
  return "AWS::ElasticLoadBalancingV2::TargetGroup"
 def get_physical_resource_id(self,attribute=hJBnv,**kwargs):
  return self.props.get("TargetGroupArn")
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("elbv2")
  group_name=self.resolve_refs_recursively(stack_name,self.props.get("Name"),resources)
  result=client.describe_target_groups()["TargetGroups"]
  result=[g for g in result if g.get("TargetGroupName")==group_name]
  return(result or[hJBnv])[0]
 @hJBnW
 def get_deploy_templates(cls):
  def register_targets(resource_id,resources,resource_type,func,stack_name,*args,**kwargs):
   resource=cls(resources[resource_id])
   resource.fetch_and_update_state(stack_name,resources)
   props=resource.props
   targets=props.get("Targets")
   if targets:
    client=aws_stack.connect_to_service("elbv2")
    group_arn=props.get("TargetGroupArn")
    client.register_targets(TargetGroupArn=group_arn,Targets=targets)
  def modify_target_group_attributes(resource_id,resources,*args,**kwargs):
   resource=cls(resources[resource_id])
   props=resource.props
   attrs=props.get("TargetGroupAttributes")
   if attrs:
    client=aws_stack.connect_to_service("elbv2")
    group_arn=props.get("TargetGroupArn")
    for attr in attrs:
     attr["Value"]=hJBnE(attr.get("Value",""))
    client.modify_target_group_attributes(TargetGroupArn=group_arn,Attributes=attrs)
  param_names=["Name","Protocol","ProtocolVersion","Port","VpcId","HealthCheckProtocol","HealthCheckPort","HealthCheckEnabled","HealthCheckPath","HealthCheckIntervalSeconds","HealthCheckTimeoutSeconds","HealthyThresholdCount","UnhealthyThresholdCount","Matcher","TargetType","Tags"]
  return{"create":[{"function":"create_target_group","parameters":param_names},{"function":register_targets},{"function":modify_target_group_attributes}]}
class ELBV2ListenerRule(GenericBaseModel):
 @hJBnk
 def cloudformation_type():
  return "AWS::ElasticLoadBalancingV2::ListenerRule"
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("RuleArn")
 def fetch_state(self,stack_name,resources):
  def actions_match(rule):
   def _matches(action,_action):
    auth_config=action.get("AuthenticateCognitoConfig",{})
    for f in["Type","Order"]:
     if action.get(f)!=_action.get(f):
      return hJBnQ
    _auth_config=_action.get("AuthenticateCognitoConfig",{})
    for f in["UserPoolArn","UserPoolClientId","UserPoolDomain"]:
     if auth_config.get(f)!=_auth_config.get(f):
      return hJBnQ
    return hJBnP
   actions1=rule.get("Actions",[])
   return items_equivalent(actions,actions1,_matches)
  def conditions_match(rule):
   def _matches(cond,_cond):
    cond_values_wrapper=([cond.get(a)for a in candidate_attrs if cond.get(a)]+[cond])[0]
    if cond.get("Field")!=_cond.get("Field"):
     return hJBnQ
    for cand in candidate_attrs:
     _cond=_cond.get(cand)or _cond
    cond_values=cond_values_wrapper.get("Values",[])
    if cond_values!=_cond.get("Values"):
     return hJBnQ
    return hJBnP
   conditions1=rule.get("Conditions",[])
   candidate_attrs=["HostHeaderConfig","HttpHeaderConfig","HttpRequestMethodConfig","PathPatternConfig","QueryStringConfig","SourceIpConfig"]
   return items_equivalent(conditions,conditions1,_matches)
  client=aws_stack.connect_to_service("elbv2")
  rp=self.props
  conditions=rp.get("Conditions",[])
  actions=rp.get("Actions",[])
  list_arn=self.resolve_refs_recursively(stack_name,rp.get("ListenerArn"),resources)
  result=client.describe_rules(ListenerArn=list_arn)["Rules"]
  result=[r for r in result if actions_match(r)and conditions_match(r)]
  return(result or[hJBnv])[0]
 @hJBnk
 def get_deploy_templates():
  return{"create":{"function":"create_rule"}}
class ELBV2Listener(GenericBaseModel):
 @hJBnk
 def cloudformation_type():
  return "AWS::ElasticLoadBalancingV2::Listener"
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("ListenerArn")
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("elbv2")
  props=self.props
  lb_arn=self.resolve_refs_recursively(stack_name,props.get("LoadBalancerArn"),resources)
  proto=self.resolve_refs_recursively(stack_name,props.get("Protocol"),resources)
  port=self.resolve_refs_recursively(stack_name,props.get("Port"),resources)
  result=client.describe_listeners(LoadBalancerArn=lb_arn)["Listeners"]
  port_candidates=[hJBnE(config.EDGE_PORT),hJBnE(config.EDGE_PORT_HTTP)]
  result_filtered=[ls for ls in result if ls.get("LoadBalancerArn")==lb_arn and(hJBnE(port or "")in[hJBnE(ls.get("Port")),""]or hJBnE(ls.get("Port"))in port_candidates)and ls.get("Protocol")==proto]
  if result and not result_filtered:
   LOG.debug("No matching entry when filtering ELBv2 listeners %s for props %s"%(result,props))
  return(result_filtered or[hJBnv])[0]
 @hJBnk
 def get_deploy_templates():
  def create_params(params,**kwargs):
   result=clone(params)
   for action in result.get("DefaultActions",[]):
    config=action.get("RedirectConfig",{})
    config["StatusCode"]=config.get("StatusCode")or "HTTP_301"
    config["Port"]=config.get("Port")and hJBnE(config["Port"])
   return result
  return{"create":{"function":"create_listener","parameters":create_params}}
class ELBV2LoadBalancer(GenericBaseModel):
 @hJBnk
 def cloudformation_type():
  return "AWS::ElasticLoadBalancingV2::LoadBalancer"
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("LoadBalancerArn")
 def fetch_state(self,stack_name,resources):
  bal_name=self.props.get("Name")
  bal_name=self.resolve_refs_recursively(stack_name,bal_name,resources)
  client=aws_stack.connect_to_service("elbv2")
  result=client.describe_load_balancers()["LoadBalancers"]
  result=[bal for bal in result if bal["LoadBalancerName"]==bal_name]
  return(result or[hJBnv])[0]
 def get_cfn_attribute(self,attribute_name):
  if attribute_name=="DNSName":
   lb_arn=self.props.get("LoadBalancerArn")
   return "%s.elb.%s"%(lb_arn.split("/")[2],ext_config.RESOURCES_BASE_DOMAIN_NAME)
  return hJBno(ELBV2LoadBalancer,self).get_cfn_attribute(attribute_name)
 @hJBnk
 def get_deploy_templates():
  def get_elbv2_loadbalancer_params(params,**kwargs):
   result=clone(params)
   result.pop("LoadBalancerAttributes",[])
   result["Name"]=result.get("Name")or kwargs.get("resource_id")
   return result
  def get_elbv2_loadbalancer_attrs_params(params,**kwargs):
   lb_name=params.get("Name")or kwargs.get("resource_id")
   result=aws_stack.connect_to_service("elbv2").describe_load_balancers()["LoadBalancers"]
   result=[bal for bal in result if bal["LoadBalancerName"]==lb_name]
   result={"LoadBalancerArn":result and result[0]["LoadBalancerArn"]or lb_name,"Attributes":params.get("LoadBalancerAttributes",[])}
   return result
  return{"create":[{"function":"create_load_balancer","parameters":get_elbv2_loadbalancer_params},{"function":"modify_load_balancer_attributes","parameters":get_elbv2_loadbalancer_attrs_params}]}
# Created by pyminifier (https://github.com/liftoff/pyminifier)



__all__ = [  # noqa: F405
    "amplify",
    "apigateway",
    "apigatewayv2",
    "applicationautoscaling",
    "appsync",
    "awslambda",
    "backup",
    "cloudfront",
    "cloudtrail",
    "cognito",
    "custom",
    "ec2",
    "ecr",
    "ecs",
    "elasticache",
    "elasticloadbalancingv2",
    "glue",
    "iot",
    "iotanalytics",
    "kinesisanalytics",
    "msk",
    "rds",
    "redshift",
    "route53",
    "servicediscovery",
    "ses",
]

# leave this star import here, so we can "import ...cloudformation.models" inside patch functions (not module level)
from localstack_ext.services.cloudformation.models import *



from localstack.services.cloudformation.service_models import GenericBaseModel
JkiKu=staticmethod
JkiKN=None
from localstack.utils.aws import aws_stack
from localstack.utils.common import short_uid
class RedshiftClusterParameterGroup(GenericBaseModel):
 @JkiKu
 def cloudformation_type():
  return "AWS::Redshift::ClusterParameterGroup"
 def get_physical_resource_id(self,attribute,**kwargs):
  return self._get_name()
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("redshift")
  group_name=self.resolve_refs_recursively(stack_name,self._get_name(),resources)
  groups=client.describe_cluster_parameter_groups(ParameterGroupName=group_name)
  groups=groups.get("ParameterGroups")
  return(groups or[JkiKN])[0]
 def _get_name(self):
  props=self.properties
  result=props["ParameterGroupName"]=(props.get("ParameterGroupName")or "cf-pg-%s"%short_uid())
  return result
 @JkiKu
 def get_deploy_templates():
  def _group_name(resource_props,resources,resource_id,*args,**kwargs):
   resource=RedshiftClusterParameterGroup(resources[resource_id])
   return resource._get_name()
  return{"create":[{"function":"create_cluster_security_group","parameters":{"ParameterGroupName":_group_name,"ParameterGroupFamily":"ParameterGroupFamily","Description":"Description","Tags":"Tags"}},{"function":"modify_cluster_parameter_group","parameters":{"ParameterGroupName":_group_name,"Parameters":"Parameters"}}],"delete":{"function":"delete_cluster_parameter_group","parameters":{"ParameterGroupName":_group_name}}}
class RedshiftClusterSecurityGroup(GenericBaseModel):
 @JkiKu
 def cloudformation_type():
  return "AWS::Redshift::ClusterSecurityGroup"
 def get_physical_resource_id(self,attribute,**kwargs):
  return self._get_name()
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("redshift")
  group_name=self.resolve_refs_recursively(stack_name,self._get_name(),resources)
  groups=client.describe_cluster_security_groups(ClusterSecurityGroupName=group_name)
  groups=groups.get("ClusterSecurityGroups")
  return(groups or[JkiKN])[0]
 def _get_name(self):
  props=self.properties
  result=props["SecurityGroupName"]=(props.get("SecurityGroupName")or "cf-sg-%s"%short_uid())
  return result
 @JkiKu
 def get_deploy_templates():
  def _group_name(resource_props,resources,resource_id,*args,**kwargs):
   resource=RedshiftClusterSecurityGroup(resources[resource_id])
   return resource._get_name()
  return{"create":{"function":"create_cluster_security_group","parameters":{"ClusterSecurityGroupName":_group_name,"Description":"Description"}},"delete":{"function":"delete_cluster_security_group","parameters":{"ClusterSecurityGroupName":_group_name}}}
class RedshiftClusterSubnetGroup(GenericBaseModel):
 @JkiKu
 def cloudformation_type():
  return "AWS::Redshift::ClusterSubnetGroup"
 def get_physical_resource_id(self,attribute,**kwargs):
  return self._get_name()
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("redshift")
  group_name=self.resolve_refs_recursively(stack_name,self._get_name(),resources)
  groups=client.describe_cluster_subnet_groups(ClusterSubnetGroupName=group_name)["ClusterSubnetGroups"]
  return(groups or[JkiKN])[0]
 def _get_name(self):
  return "-".join(self.props["SubnetIds"])
 @JkiKu
 def get_deploy_templates():
  def _group_name(resource_props,*args,**kwargs):
   return "-".join(resource_props["SubnetIds"])
  return{"create":{"function":"create_cluster_subnet_group","parameters":{"ClusterSubnetGroupName":_group_name,"Description":"Description","SubnetIds":"SubnetIds","Tags":"Tags"}},"delete":{"function":"delete_cluster_subnet_group","parameters":{"ClusterSubnetGroupName":_group_name}}}
# Created by pyminifier (https://github.com/liftoff/pyminifier)



import inspect
BWiYp=bytes
BWiYu=None
BWiYR=isinstance
BWiYo=list
BWiYt=getattr
BWiYP=open
BWiYT=property
BWiYF=Exception
BWiYw=setattr
BWiYD=True
import os.path
import sys
import traceback
from importlib.abc import MetaPathFinder,SourceLoader
from importlib.util import spec_from_file_location
import pyaes
class DecryptionHandler:
 decryption_key:BWiYp
 def __init__(self,decryption_key:BWiYp):
  self.decryption_key=decryption_key
 def decrypt(self,content)->BWiYp:
  cipher=pyaes.AESModeOfOperationCBC(self.decryption_key,iv="\0"*16)
  decrypter=pyaes.Decrypter(cipher)
  decrypted=decrypter.feed(content)
  decrypted+=decrypter.feed()
  decrypted=decrypted.partition(b"\0")[0]
  return decrypted
class EncryptedFileFinder(MetaPathFinder):
 decryption_handler:DecryptionHandler
 def __init__(self,decryption_handler:DecryptionHandler):
  self.decryption_handler=decryption_handler
 def find_spec(self,fullname,path,target=BWiYu):
  if path and not BWiYR(path,BWiYo):
   path=BWiYo(BWiYt(path,"_path",[]))
  if not path:
   return BWiYu
  name=fullname.split(".")[-1]
  file_path=os.path.join(path[0],name+".py")
  enc=file_path+".enc"
  if not os.path.isfile(enc):
   return BWiYu
  if os.path.isfile(file_path):
   return BWiYu
  return spec_from_file_location(fullname,enc,loader=DecryptingLoader(enc,self.decryption_handler))
class DecryptingLoader(SourceLoader):
 decryption_handler:DecryptionHandler
 def __init__(self,encrypted_file,decryption_handler:DecryptionHandler):
  self.encrypted_file=encrypted_file
  self.decryption_handler=decryption_handler
 def get_filename(self,fullname):
  return self.encrypted_file
 def get_data(self,filename):
  with BWiYP(filename,"rb")as f:
   data=f.read()
  data=self.decryption_handler.decrypt(data)
  return data
def init_source_decryption(decryption_handler:DecryptionHandler):
 sys.meta_path.insert(0,EncryptedFileFinder(decryption_handler))
 patch_traceback_lines()
 patch_inspect_findsource()
def patch_traceback_lines():
 if BWiYt(traceback.FrameSummary,"_ls_patch_applied",BWiYu):
  return
 @BWiYT
 def line(self):
  try:
   return line_orig.fget(self)
  except BWiYF:
   self._line=""
   return self._line
 line_orig=traceback.FrameSummary.line
 BWiYw(traceback.FrameSummary,"line",line)
 traceback.FrameSummary._ls_patch_applied=BWiYD
def patch_inspect_findsource():
 if BWiYt(inspect,"_ls_patch_applied",BWiYu):
  return
 def findsource(*args,**kwargs):
  try:
   return findsource_orig(*args,**kwargs)
  except BWiYF:
   return[],0
 findsource_orig=inspect.findsource
 inspect.findsource=findsource
 inspect._ls_patch_applied=BWiYD
# Created by pyminifier (https://github.com/liftoff/pyminifier)



from localstack.services.cloudformation.deployment_utils import lambda_keys_to_lower
JEXPf=staticmethod
JEXPM=None
JEXPi=super
JEXPR=str
from localstack.services.cloudformation.service_models import GenericBaseModel
from localstack.utils.aws import aws_stack
from localstack_ext.services.cloudformation.service_models import(lambda_add_tags,lambda_convert_types)
class ECSCluster(GenericBaseModel):
 @JEXPf
 def cloudformation_type():
  return "AWS::ECS::Cluster"
 def get_physical_resource_id(self,attribute,**kwargs):
  if attribute=="Arn":
   return self.props.get("clusterArn")
  return self.props.get("ClusterName")
 def fetch_state(self,stack_name,resources):
  props=self.props
  cluster=self.resolve_refs_recursively(stack_name,props.get("ClusterName"),resources)
  ecs_client=aws_stack.connect_to_service("ecs")
  result=ecs_client.describe_clusters(clusters=[cluster])["clusters"]
  return(result or[JEXPM])[0]
 @JEXPf
 def get_deploy_templates():
  return{"create":{"function":"create_cluster","parameters":lambda_keys_to_lower()}}
class ECSService(GenericBaseModel):
 @JEXPf
 def cloudformation_type():
  return "AWS::ECS::Service"
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("serviceArn")
 def fetch_state(self,stack_name,resources):
  props=self.props
  cluster=self.resolve_refs_recursively(stack_name,props.get("Cluster"),resources)
  service_name=self.resolve_refs_recursively(stack_name,props.get("ServiceName"),resources)
  ecs_client=aws_stack.connect_to_service("ecs")
  result=ecs_client.describe_services(cluster=cluster,services=[service_name])["services"]
  return(result or[JEXPM])[0]
 def get_cfn_attribute(self,attribute_name):
  if attribute_name=="Name":
   return self.props.get("ServiceName")
  return JEXPi(ECSService,self).get_cfn_attribute(attribute_name)
 @JEXPf
 def get_deploy_templates():
  return{"create":{"function":"create_service","parameters":lambda_add_tags(lambda_keys_to_lower())}}
class ECSTaskDefinition(GenericBaseModel):
 @JEXPf
 def cloudformation_type():
  return "AWS::ECS::TaskDefinition"
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("taskDefinitionArn")
 def fetch_state(self,stack_name,resources):
  task_def=self.props.get("Family")
  task_def=self.resolve_refs_recursively(stack_name,task_def,resources)
  ecs_client=aws_stack.connect_to_service("ecs")
  task_defs=ecs_client.list_task_definitions(familyPrefix=task_def)["taskDefinitionArns"]
  return task_defs and{"taskDefinitionArn":task_defs[0]}or JEXPM
 @JEXPf
 def get_deploy_templates():
  return{"create":{"function":"register_task_definition","parameters":lambda_convert_types(lambda_keys_to_lower(),{".memory":JEXPR,".cpu":JEXPR})}}
# Created by pyminifier (https://github.com/liftoff/pyminifier)



import logging
uOcLN=bool
uOcLl=True
uOcLi=False
uOcLg=Exception
uOcLS=any
uOcLj=None
import os
from localstack import config as localstack_config
from localstack.runtime import hooks
from localstack.utils import common
from localstack.utils.bootstrap import(API_DEPENDENCIES,LocalstackContainer,get_enabled_apis,is_api_enabled)
from localstack.utils.common import is_mac_os
from localstack.utils.docker_utils import VolumeBind
from localstack_ext import config as config_ext
from localstack_ext.bootstrap import install,licensing,local_daemon
LOG=logging.getLogger(__name__)
EXTERNAL_PORT_APIS=("apigateway","apigatewayv2","athena","cloudfront","codecommit","ecs","ecr","elasticache","mediastore","rds","transfer","kafka","neptune","azure")
API_DEPENDENCIES.update({"amplify":["s3","appsync","cognito"],"apigateway":["apigatewayv2"],"athena":["emr"],"docdb":["rds"],"ecs":["ecr"],"elasticache":["ec2"],"elb":["elbv2"],"emr":["athena","s3"],"glacier":["s3"],"glue":["rds"],"iot":["iot-analytics","iot-data","iotwireless"],"kinesisanalytics":["kinesis","dynamodb"],"neptune":["rds"],"rds":["rds-data"],"redshift":["redshift-data"],"timestream":["timestream-write","timestream-query"],"transfer":["s3"]})
get_enabled_apis.cache_clear()
def api_key_configured()->uOcLN:
 return uOcLl if os.environ.get("LOCALSTACK_API_KEY")else uOcLi
def modify_edge_port_config():
 if os.environ.get("EDGE_PORT")and not localstack_config.EDGE_PORT_HTTP:
  LOG.warning(("!! Configuring EDGE_PORT={p} without setting EDGE_PORT_HTTP may lead "+"to issues; better leave the defaults, or set EDGE_PORT=443 and EDGE_PORT_HTTP={p}").format(p=localstack_config.EDGE_PORT))
 else:
  port=localstack_config.EDGE_PORT
  localstack_config.EDGE_PORT=443
  localstack_config.EDGE_PORT_HTTP=port
@hooks.install(should_load=api_key_configured)
def install_pro_libs():
 install.install_libs()
@hooks.prepare_host(priority=100,should_load=api_key_configured)
def activate_pro_key_on_host():
 with licensing.prepare_environment():
  LOG.debug("pro activation done")
@hooks.prepare_host(should_load=api_key_configured)
def create_dns_forward():
 try:
  from localstack_ext.services import dns_server
  dns_server.setup_network_configuration()
 except uOcLg as e:
  LOG.warning("Unable to start DNS: %s"%e)
@hooks.prepare_host(should_load=api_key_configured)
def start_ec2_daemon():
 try:
  if is_api_enabled("ec2")and is_mac_os():
   LOG.debug("Starting EC2 daemon...")
   local_daemon.start_in_background()
 except uOcLg as e:
  LOG.warning("Unable to start local daemon process: %s"%e)
@hooks.configure_localstack_container(priority=10,should_load=api_key_configured)
def configure_pro_container(container:LocalstackContainer):
 try:
  from localstack_ext.services import dns_server
  docker_flags=[]
  if config_ext.use_custom_dns():
   if not common.is_port_open(dns_server.DNS_PORT,protocols="tcp"):
    docker_flags+=["-p","{a}:{p}:{p}".format(a=config_ext.DNS_ADDRESS,p=dns_server.DNS_PORT)]
   if not common.is_port_open(dns_server.DNS_PORT,protocols="udp"):
    docker_flags+=["-p","{a}:{p}:{p}/udp".format(a=config_ext.DNS_ADDRESS,p=dns_server.DNS_PORT)]
  if uOcLS([is_api_enabled(api)for api in EXTERNAL_PORT_APIS]):
   docker_flags+=["-p","{start}-{end}:{start}-{end}".format(start=config_ext.SERVICE_INSTANCES_PORTS_START,end=config_ext.SERVICE_INSTANCES_PORTS_END)]
  container.additional_flags.extend(docker_flags)
 except uOcLg as e:
  LOG.warning("failed to configure DNS: %s",e)
 modify_edge_port_config()
 if is_api_enabled("eks"):
  kube_config=os.path.expanduser("~/.kube/config")
  if os.path.exists(kube_config):
   container.volumes.add(VolumeBind(kube_config,"/root/.kube/config"))
 if is_api_enabled("azure"):
  container.ports.add(5671)
 if os.environ.get("AZURE"):
  container.ports.add(config_ext.PORT_AZURE)
@hooks.on_infra_start(should_load=api_key_configured)
def setup_pro_infra():
 _setup_logging()
 install.setup_ssl_cert()
 modify_edge_port_config()
 with licensing.prepare_environment():
  try:
   from localstack_ext.services import dns_server
   dns_server.setup_network_configuration()
  except uOcLg as e:
   LOG.warning("error setting up dns server: %s",e)
  try:
   from localstack_ext.bootstrap.dashboard import dashboard_extended
   from localstack_ext.services import edge
   from localstack_ext.utils import persistence as persistence_ext
   from localstack_ext.utils.aws import aws_utils
   persistence_ext.enable_extended_persistence()
   dashboard_extended.patch_dashboard()
   edge.patch_start_edge()
   patch_start_infra()
   aws_utils.patch_aws_utils()
   set_default_providers_to_pro()
  except uOcLg as e:
   if LOG.isEnabledFor(level=logging.DEBUG):
    LOG.exception("error enabling pro code")
   else:
    LOG.error("error enabling pro code: %s",e)
def set_default_providers_to_pro():
 from localstack.services.plugins import SERVICE_PLUGINS
 pro_services=SERVICE_PLUGINS.apis_with_provider("pro")
 localstack_config.SERVICE_PROVIDER_CONFIG.bulk_set_provider_if_not_exists(pro_services,"pro")
 eager_services=["azure"]
 for service in eager_services:
  if not is_api_enabled(service):
   continue
  try:
   LOG.debug("loading service plugin for %s",service)
   SERVICE_PLUGINS.get_service_container(service).start()
  except uOcLg as e:
   LOG.error("error while loading service %s: %s",service,e)
def patch_start_infra():
 from localstack.services import infra
 try:
  from localstack_ext.utils.aws.metadata_service import start_metadata_service
 except uOcLg:
  start_metadata_service=uOcLj
 def do_start_infra(asynchronous,apis,is_in_docker,*args,**kwargs):
  if common.in_docker():
   try:
    start_metadata_service and start_metadata_service()
   except uOcLg:
    pass
  enforce_before=config_ext.ENFORCE_IAM
  try:
   config_ext.ENFORCE_IAM=uOcLi
   return do_start_infra_orig(asynchronous,apis,is_in_docker,*args,**kwargs)
  finally:
   config_ext.ENFORCE_IAM=enforce_before
 do_start_infra_orig=infra.do_start_infra
 infra.do_start_infra=do_start_infra
def _setup_logging():
 log_level=logging.DEBUG if localstack_config.DEBUG else logging.INFO
 logging.getLogger("localstack_ext").setLevel(log_level)
 logging.getLogger("botocore").setLevel(logging.INFO)
 logging.getLogger("kubernetes").setLevel(logging.INFO)
 logging.getLogger("pyftpdlib").setLevel(logging.INFO)
 logging.getLogger("pyhive").setLevel(logging.INFO)
 logging.getLogger("websockets").setLevel(logging.INFO)
 logging.getLogger("asyncio").setLevel(logging.INFO)
 logging.getLogger("hpack").setLevel(logging.INFO)
 logging.getLogger("jnius.reflect").setLevel(logging.INFO)
 logging.getLogger("dulwich").setLevel(logging.ERROR)
 logging.getLogger("kazoo").setLevel(logging.ERROR)
 logging.getLogger("postgresql_proxy").setLevel(logging.WARNING)
 logging.getLogger("intercept").setLevel(logging.WARNING)
 logging.getLogger("root").setLevel(logging.WARNING)
 logging.getLogger("").setLevel(logging.WARNING)
# Created by pyminifier (https://github.com/liftoff/pyminifier)



#!/usr/bin/env python
from setuptools import setup

entry_points = {
    "localstack.installer": ["ext=localstack_ext.bootstrap.install:ExtInstallerRepository"],
    "localstack.hooks.prepare_host": [
        "activate_pro_key_on_host=localstack_ext.plugins:activate_pro_key_on_host",
        "create_dns_forward=localstack_ext.plugins:create_dns_forward",
        "start_ec2_container=localstack_ext.plugins:start_ec2_daemon",
    ],
    "localstack.hooks.configure_localstack_container": [
        "configure_pro_container=localstack_ext.plugins:configure_pro_container"
    ],
    "localstack.hooks.install": ["install_pro_libs=localstack_ext.plugins:install_pro_libs"],
    "localstack.hooks.on_infra_start": ["setup_pro_infra=localstack_ext.plugins:setup_pro_infra"],
    "localstack.aws.provider": [
        "amplify:pro=localstack_ext.services.providers:amplify",
        "apigateway:pro=localstack_ext.services.providers:apigateway",
        "appconfig:pro=localstack_ext.services.providers:appconfig",
        "application-autoscaling:pro=localstack_ext.services.providers:application_autoscaling",
        "appsync:pro=localstack_ext.services.providers:appsync",
        "athena:pro=localstack_ext.services.providers:athena",
        "autoscaling:pro=localstack_ext.services.providers:autoscaling",
        "lambda:pro=localstack_ext.services.providers:awslambda",
        "azure:pro=localstack_ext.services.providers:azure",
        "backup:pro=localstack_ext.services.providers:backup",
        "batch:pro=localstack_ext.services.providers:batch",
        "ce:pro=localstack_ext.services.providers:ce",
        "cloudformation:pro=localstack_ext.services.providers:cloudformation",
        "cloudfront:pro=localstack_ext.services.providers:cloudfront",
        "cloudtrail:pro=localstack_ext.services.providers:cloudtrail",
        "codecommit:pro=localstack_ext.services.providers:codecommit",
        "cognito-identity:pro=localstack_ext.services.providers:cognito_identity",
        "cognito-idp:pro=localstack_ext.services.providers:cognito_idp",
        "docdb:pro=localstack_ext.services.providers:docdb",
        "dynamodb:pro=localstack_ext.services.providers:dynamodb",
        "ec2:pro=localstack_ext.services.providers:ec2",
        "ecr:pro=localstack_ext.services.providers:ecr",
        "ecs:pro=localstack_ext.services.providers:ecs",
        "efs:pro=localstack_ext.services.providers:efs",
        "eks:pro=localstack_ext.services.providers:eks",
        "elasticache:pro=localstack_ext.services.providers:elasticache",
        "elasticbeanstalk:pro=localstack_ext.services.providers:elasticbeanstalk",
        "elb:pro=localstack_ext.services.providers:elb",
        "elbv2:pro=localstack_ext.services.providers:elbv2",
        "emr:pro=localstack_ext.services.providers:emr",
        "events:pro=localstack_ext.services.providers:events",
        "glacier:pro=localstack_ext.services.providers:glacier",
        "glue:pro=localstack_ext.services.providers:glue",
        "iam:pro=localstack_ext.services.providers:iam",
        "iot:pro=localstack_ext.services.providers:iot",
        "iot-analytics:pro=localstack_ext.services.providers:iot_analytics",
        "iot-data:pro=localstack_ext.services.providers:iot_data",
        "kafka:pro=localstack_ext.services.providers:kafka",
        "kinesisanalytics:pro=localstack_ext.services.providers:kinesisanalytics",
        "kms:pro=localstack_ext.services.providers:kms",
        "lakeformation:pro=localstack_ext.services.providers:lakeformation",
        "mediastore:pro=localstack_ext.services.providers:mediastore",
        "neptune:pro=localstack_ext.services.providers:neptune",
        "organizations:pro=localstack_ext.services.providers:organizations",
        "qldb:pro=localstack_ext.services.providers:qldb",
        "rds:pro=localstack_ext.services.providers:rds",
        "rds-data:pro=localstack_ext.services.providers:rds_data",
        "redshift:pro=localstack_ext.services.providers:redshift",
        "redshift-data:pro=localstack_ext.services.providers:redshift_data",
        "route53:pro=localstack_ext.services.providers:route53",
        "s3:pro=localstack_ext.services.providers:s3",
        "sagemaker:pro=localstack_ext.services.providers:sagemaker",
        "secretsmanager:pro=localstack_ext.services.providers:secretsmanager",
        "serverlessrepo:pro=localstack_ext.services.providers:serverlessrepo",
        "servicediscovery:pro=localstack_ext.services.providers:servicediscovery",
        "ses:pro=localstack_ext.services.providers:ses",
        "sns:pro=localstack_ext.services.providers:sns",
        "sqs:pro=localstack_ext.services.providers:sqs",
        "sts:pro=localstack_ext.services.providers:sts",
        "timestream:pro=localstack_ext.services.providers:timestream",
        "timestream-query:pro=localstack_ext.services.providers:timestream_query",
        "timestream-write:pro=localstack_ext.services.providers:timestream_write",
        "transfer:pro=localstack_ext.services.providers:transfer",
        "xray:pro=localstack_ext.services.providers:xray",
    ],
    "localstack.plugins.cli": ["pro=localstack_ext.cli.localstack:ProCliPlugin"],
}

setup(entry_points=entry_points)



import glob
HKNCO=True
HKNCX=dict
HKNCo=open
HKNCR=False
HKNCl=Exception
HKNCM=int
import json
import logging
import os
import subprocess
import sys
import time
from ftplib import FTP
from localstack import config as localstack_config
from localstack.services.generic_proxy import GenericProxy
from localstack.utils.aws import aws_stack
from localstack.utils.common import(TMP_THREADS,ShellCommandThread,load_file,new_tmp_dir,save_file,wait_for_port_open)
from localstack.utils.run import FuncThread
from pyftpdlib.authorizers import DummyAuthorizer
from pyftpdlib.handlers import FTPHandler,TLS_FTPHandler
from pyftpdlib.servers import FTPServer
from localstack_ext import config as ext_config
LOG=logging.getLogger(__name__)
ROOT_USER=("root","pass123")
FTP_USER_DEFAULT_PASSWD="12345"
FTP_USER_PERMISSIONS="elradfmwMT"
FTP_PASSIVE_PORTS=[ext_config.SERVICE_INSTANCES_PORTS_END-2,ext_config.SERVICE_INSTANCES_PORTS_END-1,ext_config.SERVICE_INSTANCES_PORTS_END]
USE_SUBPROCESS=HKNCO
DIRECTORY_MAPPING={}
DIRECTORY_MAPPING_FILE="<data_dir>/ftp.user.dir.mapping.json"
def get_dir_mapping_key(username,server_port):
 return "{}:{}".format(username,server_port)
def apply_patches():
 extended_proto_cmds=TLS_FTPHandler.proto_cmds.copy()
 extended_proto_cmds.update({"SITE ADDUSER":HKNCX(perm="m",auth=HKNCO,arg=HKNCO,help="Syntax: SITE <SP> ADDUSER USERNAME PASSWORD HOME PRIVS <SP>.")})
 def _on_file_received(self,file_path):
  key=get_dir_mapping_key(self.username,self.server.address[1])
  mapping=get_directory_mapping()
  configuration=mapping.get(key,{})
  if not configuration:
   return
  bucket=configuration["HomeDirectory"]
  user_dir=configuration["UserDirectory"]
  key=file_path.replace("{}/".format(user_dir),"")
  with HKNCo(file_path)as f:
   s3_client=aws_stack.connect_to_service("s3")
   s3_client.put_object(Bucket=bucket,Key=key,Body=f.read())
   LOG.info("Received file via FTP -- target: s3://{}/{}".format(bucket,key))
 def _site_adduser(self,line):
  user,passwd,user_dir,perm=line.split(" ")[1:]
  self.authorizer.add_user(user,passwd,user_dir,perm)
  self.respond("201 Add User OK.")
 FTPHandler.proto_cmds=extended_proto_cmds
 TLS_FTPHandler.proto_cmds=extended_proto_cmds
 FTPHandler.on_file_received=_on_file_received
 FTPHandler.ftp_SITE_ADDUSER=_site_adduser
 TLS_FTPHandler.on_file_received=_on_file_received
 TLS_FTPHandler.ftp_SITE_ADDUSER=_site_adduser
def add_ftp_user(user,server_port):
 ftp=FTP()
 ftp.connect(localstack_config.LOCALSTACK_HOSTNAME,port=server_port)
 ftp.login(ROOT_USER[0],ROOT_USER[1])
 user_dir=new_tmp_dir()
 ftp.sendcmd("SITE ADDUSER  {} {} {} {}".format(user.username,FTP_USER_DEFAULT_PASSWD,user_dir,FTP_USER_PERMISSIONS))
 ftp.quit()
 dir_mapping_key=get_dir_mapping_key(user.username,server_port)
 mapping=user.get_directory_configuration()
 mapping.update({"UserDirectory":user_dir})
 set_directory_mapping(dir_mapping_key,mapping)
def update_ftp_user(user,server_port):
 dir_mapping_key=get_dir_mapping_key(user.username,server_port)
 set_directory_mapping(dir_mapping_key,user.get_directory_configuration())
def start_ftp(port):
 if USE_SUBPROCESS:
  services=os.environ.get("SERVICES","")
  if services and "s3" not in services:
   services+=",s3"
  pythonpath=os.environ.get("PYTHONPATH")or ""
  if os.getcwd()not in pythonpath.split(":"):
   pythonpath="%s:%s"%(os.getcwd(),pythonpath)
  paths=glob.glob("%s/.venv/lib/python*/site-packages"%os.getcwd())
  if paths:
   pythonpath+=":%s"%":".join(paths)
  env_vars={"SERVICES":services,"PYTHONPATH":pythonpath}
  cmd="%s %s %s"%(sys.executable,__file__,port)
  thread=ShellCommandThread(cmd,outfile=subprocess.PIPE,env_vars=env_vars,quiet=HKNCR)
  thread.start()
  time.sleep(2)
 else:
  thread=do_start_ftp(port,asynchronous=HKNCO)
 time.sleep(2)
 wait_for_port_open(port,retries=10,sleep_time=1.5)
 TMP_THREADS.append(thread)
 return thread
def do_start_ftp(port,asynchronous=HKNCO):
 LOG.info("Starting (S)FTP server on port %s..."%port)
 apply_patches()
 authorizer=DummyAuthorizer()
 user_dir=new_tmp_dir()
 authorizer.add_user(ROOT_USER[0],ROOT_USER[1],user_dir,perm=FTP_USER_PERMISSIONS)
 anonymous_dir=new_tmp_dir()
 authorizer.add_anonymous(anonymous_dir)
 handler=TLS_FTPHandler
 combined_file,_,_=GenericProxy.create_ssl_cert()
 handler.certfile=combined_file
 handler.authorizer=authorizer
 handler.passive_ports=FTP_PASSIVE_PORTS
 handler.masquerade_address=ext_config.LOCALHOST_IP
 def do_run(*args):
  try:
   server=FTPServer(("0.0.0.0",port),handler)
   server.serve_forever()
  except HKNCl as e:
   LOG.info("Unable to run FTP server on port %s: %s"%(port,e))
   raise
 if asynchronous:
  t=FuncThread(do_run)
  t.start()
  return t
 return do_run()
def get_directory_mapping():
 result=DIRECTORY_MAPPING
 if USE_SUBPROCESS:
  dir_file=get_directory_mapping_file()
  result=json.loads(load_file(dir_file)or "{}")
 return result
def set_directory_mapping(key,value):
 mapping=get_directory_mapping()
 mapping[key]=value
 if USE_SUBPROCESS:
  dir_file=get_directory_mapping_file()
  save_file(dir_file,json.dumps(mapping))
 return value
def get_directory_mapping_file():
 return DIRECTORY_MAPPING_FILE.replace("<data_dir>",localstack_config.dirs.tmp)
def main():
 do_start_ftp(HKNCM(sys.argv[1]),asynchronous=HKNCR)
if __name__=="__main__":
 main()
# Created by pyminifier (https://github.com/liftoff/pyminifier)



import hashlib
ktoLx=str
ktoLe=hex
ktoLs=None
ktoLn=open
ktoLD=Exception
ktoLN=int
ktoLu=map
ktoLF=isinstance
import logging
import os
import random
import zipfile
from typing import Optional
from localstack.utils.common import new_tmp_dir,rm_rf
from localstack_ext.bootstrap.cpvcs.models import Revision,Version
from localstack_ext.bootstrap.cpvcs.utils.common import config_context
from localstack_ext.bootstrap.state_utils import API_STATES_DIR,api_states_traverse
LOG=logging.getLogger(__name__)
def random_hash()->ktoLx:
 return ktoLe(random.getrandbits(160))
def compute_file_hash(file_path:ktoLx,accum=ktoLs)->Optional[ktoLx]:
 try:
  with ktoLn(file_path,"rb")as fp:
   if accum:
    accum.update(fp.read())
   else:
    return hashlib.sha1(fp.read()).hexdigest()
 except ktoLD as e:
  LOG.warning(f"Failed to open file and compute hash for file at {file_path}: {e}")
def compute_version_archive_hash(version_no:ktoLN,state_archive:ktoLx)->ktoLx:
 def _compute_file_hash_func(**kwargs):
  dir_name=kwargs.get("dir_name")
  file_name=kwargs.get("fname")
  accum=kwargs.get("mutables")[0]
  file_path=os.path.join(dir_name,file_name)
  compute_file_hash(file_path,accum)
 tmp_state_dir=new_tmp_dir()
 with zipfile.ZipFile(state_archive)as archive:
  archive.extractall(tmp_state_dir)
 tmp_state_archive_api_states=os.path.join(tmp_state_dir,API_STATES_DIR)
 m=hashlib.sha1()
 api_states_traverse(tmp_state_archive_api_states,_compute_file_hash_func,[m])
 m.update(ktoLx(version_no).encode("utf-8"))
 rm_rf(tmp_state_dir)
 return m.hexdigest()
def compute_revision_hash(cpvcs_node:Revision)->ktoLx:
 if not cpvcs_node.state_files:
  return random_hash()
 state_file_keys=ktoLu(lambda state_file:state_file.hash_ref,cpvcs_node.state_files)
 m=hashlib.sha1()
 for key in state_file_keys:
  try:
   with ktoLn(os.path.join(config_context.get_obj_store_path(),key),"rb")as fp:
    m.update(fp.read())
  except ktoLD as e:
   LOG.warning(f"Failed to open file and compute hash for {key}: {e}")
 if ktoLF(cpvcs_node,Revision):
  m.update(cpvcs_node.rid.encode("utf-8"))
  m.update(ktoLx(cpvcs_node.revision_number).encode("utf-8"))
 elif ktoLF(cpvcs_node,Version):
  m.update(ktoLx(cpvcs_node.version_number).encode("utf-8"))
 return m.hexdigest()
# Created by pyminifier (https://github.com/liftoff/pyminifier)



from localstack.services.cloudformation.service_models import REF_ATTRS,GenericBaseModel
vBMpP=staticmethod
vBMpl=super
from localstack.utils.aws import aws_stack
from localstack.utils.common import clone
class GlueTrigger(GenericBaseModel):
 @vBMpP
 def cloudformation_type():
  return "AWS::Glue::Trigger"
 def get_cfn_attribute(self,attribute_name):
  if attribute_name in REF_ATTRS:
   return self.props.get("Name")
  return vBMpl(GlueTrigger,self).get_cfn_attribute(attribute_name)
class GlueWorkflow(GenericBaseModel):
 @vBMpP
 def cloudformation_type():
  return "AWS::Glue::Workflow"
 def get_cfn_attribute(self,attribute_name):
  if attribute_name in REF_ATTRS:
   return self.props.get("Name")
  return vBMpl(GlueWorkflow,self).get_cfn_attribute(attribute_name)
class GlueJob(GenericBaseModel):
 @vBMpP
 def cloudformation_type():
  return "AWS::Glue::Job"
 def get_cfn_attribute(self,attribute_name):
  if attribute_name in REF_ATTRS:
   return self.props["Name"]
  return vBMpl(GlueJob,self).get_cfn_attribute(attribute_name)
class GlueCrawler(GenericBaseModel):
 @vBMpP
 def cloudformation_type():
  return "AWS::Glue::Crawler"
 def get_cfn_attribute(self,attribute_name):
  if attribute_name in REF_ATTRS:
   return self.props["Name"]
  return vBMpl(GlueCrawler,self).get_cfn_attribute(attribute_name)
 def fetch_state(self,stack_name,resources):
  crawler_name=self.props.get("Name")or self.resource_id
  crawler_name=self.resolve_refs_recursively(stack_name,crawler_name,resources)
  client=aws_stack.connect_to_service("glue")
  return client.get_crawler(Name=crawler_name)["Crawler"]
 @vBMpP
 def get_deploy_templates():
  def lambda_get_crawler_config(params,**kwargs):
   config=clone(params)
   if "Schedule" in config:
    config["Schedule"]=config["Schedule"]["ScheduleExpression"]
   config["Role"]=config.get("Role")or "_unknown_"
   return config
  return{"create":{"function":"create_crawler","parameters":lambda_get_crawler_config}}
class GlueDatabase(GenericBaseModel):
 @vBMpP
 def cloudformation_type():
  return "AWS::Glue::Database"
 def get_cfn_attribute(self,attribute_name):
  if attribute_name in REF_ATTRS:
   return self.props.get("DatabaseInput",{}).get("Name")
  return vBMpl(GlueDatabase,self).get_cfn_attribute(attribute_name)
class GlueClassifier(GenericBaseModel):
 @vBMpP
 def cloudformation_type():
  return "AWS::Glue::Classifier"
 def get_cfn_attribute(self,attribute_name):
  if attribute_name in REF_ATTRS:
   result=(self.props.get("GrokClassifier",{}).get("Name")or self.props.get("CsvClassifier",{}).get("Name")or self.props.get("JsonClassifier",{}).get("Name")or self.props.get("XMLClassifier",{}).get("Name"))
   if result:
    return result
  return vBMpl(GlueCrawler,self).get_cfn_attribute(attribute_name)
class GlueTable(GenericBaseModel):
 @vBMpP
 def cloudformation_type():
  return "AWS::Glue::Table"
 def get_cfn_attribute(self,attribute_name):
  if attribute_name in REF_ATTRS:
   return self.props.get("TableInput",{}).get("Name")
  return vBMpl(GlueTable,self).get_cfn_attribute(attribute_name)
# Created by pyminifier (https://github.com/liftoff/pyminifier)



import os
jposc=Exception
jposK=hasattr
from localstack import config
from localstack.dashboard import infra as dashboard_infra
from localstack.utils.aws import aws_stack
from localstack.utils.aws.request_context import MARKER_APIGW_REQUEST_REGION,THREAD_LOCAL
from localstack.utils.bootstrap import is_api_enabled
from localstack.utils.common import run_safe,short_uid
from localstack_ext.bootstrap.aws_models import(AmplifyApp,AppSyncAPI,CloudFrontDistribution,CodeCommitRepository,ElastiCacheCluster,RDSCluster,RDSDatabase,TransferServer)
get_graph_orig=dashboard_infra.get_graph
def get_resources(fetch_func):
 try:
  result=[]
  fetch_func(result)
  return result
 except jposc:
  pass
 return[]
def get_rds_databases(name_filter,pool,env):
 if not is_api_enabled("rds"):
  return[]
 def fetch_func(result):
  client=aws_stack.connect_to_service("rds")
  dbs=client.describe_db_instances()
  for inst in dbs["DBInstances"]:
   obj=RDSDatabase(id=inst["DBInstanceArn"])
   result.append(obj)
 return get_resources(fetch_func)
def get_rds_clusters(name_filter,pool,env):
 if not is_api_enabled("rds"):
  return[]
 def fetch_func(result):
  client=aws_stack.connect_to_service("rds")
  clusters=client.describe_db_clusters()
  for cluster in clusters["DBClusters"]:
   obj=RDSCluster(id=cluster["DBClusterArn"])
   result.append(obj)
 return get_resources(fetch_func)
def get_appsync_apis(name_filter,pool,env):
 if not is_api_enabled("appsync"):
  return[]
 def fetch_func(result):
  client=aws_stack.connect_to_service("appsync")
  apis=client.list_graphql_apis()
  for api in apis["graphqlApis"]:
   obj=AppSyncAPI(id=api["apiId"])
   result.append(obj)
 return get_resources(fetch_func)
def get_amplify_apps(name_filter,pool,env):
 if not is_api_enabled("amplify"):
  return[]
 def fetch_func(result):
  client=aws_stack.connect_to_service("amplify")
  apps=client.list_apps()
  for app in apps["apps"]:
   obj=AmplifyApp(id=app["appId"])
   result.append(obj)
 return get_resources(fetch_func)
def get_elasticache_clusters(name_filter,pool,env):
 if not is_api_enabled("elasticache"):
  return[]
 def fetch_func(result):
  client=aws_stack.connect_to_service("elasticache")
  clusters=client.describe_cache_clusters()
  for cluster in clusters["CacheClusters"]:
   obj=ElastiCacheCluster(id=cluster["CacheClusterId"])
   result.append(obj)
 return get_resources(fetch_func)
def get_transfer_servers(name_filter,pool,env):
 if not is_api_enabled("transfer"):
  return[]
 def fetch_func(result):
  client=aws_stack.connect_to_service("transfer")
  servers=client.list_servers()
  for server in servers["Servers"]:
   obj=TransferServer(id=server["ServerId"])
   result.append(obj)
 return get_resources(fetch_func)
def get_cloudfront_distributions(name_filter,pool,env):
 if not is_api_enabled("cloudfront"):
  return[]
 def fetch_func(result):
  client=aws_stack.connect_to_service("cloudfront")
  distros=client.list_distributions()
  for distro in distros["DistributionList"]["Items"]:
   obj=CloudFrontDistribution(id=distro["ARN"])
   result.append(obj)
 return get_resources(fetch_func)
def get_codecommit_repos(name_filter,pool,env):
 if not is_api_enabled("codecommit"):
  return[]
 def fetch_func(result):
  client=aws_stack.connect_to_service("codecommit")
  repos=client.list_repositories()
  for repo in repos["repositories"]:
   obj=CodeCommitRepository(id=repo["repositoryId"])
   result.append(obj)
 return get_resources(fetch_func)
def get_graph(*args,**kwargs):
 os.environ["AWS_ACCESS_KEY_ID"]=os.environ.get("AWS_ACCESS_KEY_ID")or "foobar"
 os.environ["AWS_SECRET_ACCESS_KEY"]=os.environ.get("AWS_SECRET_ACCESS_KEY")or "foobar"
 if jposK(THREAD_LOCAL,"request_context"):
  region=kwargs.get("region")or config.DEFAULT_REGION
  THREAD_LOCAL.request_context.headers[MARKER_APIGW_REQUEST_REGION]=region
 result=run_safe(get_graph_orig,*args,**kwargs)or{"nodes":[],"edges":[]}
 env=kwargs.get("env")
 name_filter=kwargs.get("name_filter")
 pool={}
 node_ids={}
 databases=get_rds_databases(name_filter,pool=pool,env=env)
 rds_clusters=get_rds_clusters(name_filter,pool=pool,env=env)
 appsync_apis=get_appsync_apis(name_filter,pool=pool,env=env)
 amplify_apps=get_amplify_apps(name_filter,pool=pool,env=env)
 elasticache_clusters=get_elasticache_clusters(name_filter,pool=pool,env=env)
 transfer_servers=get_transfer_servers(name_filter,pool=pool,env=env)
 cloudfront_distributions=get_cloudfront_distributions(name_filter,pool=pool,env=env)
 codecommit_repos=get_codecommit_repos(name_filter,pool=pool,env=env)
 resources={"rds_db":databases,"rds_cluster":rds_clusters,"appsync_api":appsync_apis,"amplify_app":amplify_apps,"elasticache_cluster":elasticache_clusters,"transfer_server":transfer_servers,"cloudfront_distr":cloudfront_distributions,"codecommit_repo":codecommit_repos}
 for res_type,res_list in resources.items():
  for res in res_list:
   uid=short_uid()
   node_ids[res.id]=uid
   result["nodes"].append({"id":uid,"arn":res.id,"name":res.name(),"type":res_type})
 return result
def patch_dashboard():
 dashboard_infra.get_graph=get_graph
# Created by pyminifier (https://github.com/liftoff/pyminifier)
